/*--------------------------------------------------------------
	Concatenate PR-list files     bxcat  v1.1
		Written  by H.Goto , July 1993
		Modified by H.Goto , Apr. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1993-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>

#include	"ufilep.h"




/*------------------------------------------------------
	Message class
------------------------------------------------------*/

class MSG {
    public:
	void		usage(void);
	void		switcherr(void);
	void		readerr(char *name);
	void		writeerr(char *name);
	void		createrr(char *name);
	void		nonsupported(char *name);
	void		memoryover(void);
	void		toomanyboxes(void);
};
    	
    	
void MSG :: usage(void){
	fputs("bxcat  v1.1,  Copyright (C) 1993-1997 Hideaki Goto \n",stderr);
	fputs("Usage : bxcat  infile1 [infile2] [...] -o outfile \n",stderr);
//	fputs("		-quiet   : suppress messages \n",stderr);
}


void MSG :: switcherr(void){
	fputs("Switch error.\n",stderr);
}


void MSG :: readerr(char *name){
	fprintf(stderr,"Can't read %s.\n",name);
}


void MSG :: writeerr(char *name){
	fprintf(stderr,"Can't write %s.\n",name);
}


void MSG :: createrr(char *name){
	fprintf(stderr,"Can't create %s.\n",name);
}


void MSG :: nonsupported(char *name){
	fprintf(stderr,"Non-supported file %s.\n",name);
}


void MSG :: memoryover(void){
	fputs("Not enough memory.\n",stderr);
}


void MSG :: toomanyboxes(void){
	fputs("Too many boxes.\n",stderr);
}



/*------------------------------------------------------
	Argument / Option class
------------------------------------------------------*/

class ARGS {
    private:
	char		**avsave;
	char		*out_fname;
	int		fpstart;
	int		fpcount;
	int		fpgetcount;
    	int		sw_verbose;
    public:
	int		scanargv(int ac,char **av);
	int		SWverbose(void) {  return sw_verbose; }
	char		*infile_rewind(void){  fpgetcount = 0; }
	char		*infile_next(void);
	char		*outfile(void);
			ARGS(void);
};


ARGS :: ARGS(void){
	sw_verbose = -1;
	out_fname = 0;
}


int ARGS :: scanargv(int ac,char **av){
	int	i,err,skip;
	char	*sw;
	fpstart = fpcount = fpgetcount = err = 0;
	skip = 0;
	avsave = av;
	for( i=1 ; i < ac ; ++i ){
		if( skip ){  skip = 0;  continue;  }
		if( *( sw = av[i] ) == '-' ){
			++sw;
			if( 0 == strcmp(sw,"o") ){
				if((i+1) < ac){
					out_fname = av[i+1];
					skip = -1;
					continue;
				}
				else{  err = -1;  break; }
			}
			if( 0 == strcmp(sw,"quiet") ){
				sw_verbose = 0;
				continue;
			}
			if( 0 == strcmp(sw,"verbose") ){
				sw_verbose = -1;
				continue;
			}
			err = -1;  break;
		}
		else{	if ( fpstart == 0 )  fpstart = i;
			++fpcount;
		}
	}
	return(err);
}


char * ARGS :: infile_next(void){
	if ( fpgetcount >= fpcount )  return(0);
	return( avsave[ fpstart + (fpgetcount++) ] );
}


char * ARGS :: outfile(void){
	if ( out_fname == 0 )  return( "/dev/null" );
	return ( out_fname );
}




/*------------------------------------------------------
	Main routine
------------------------------------------------------*/


int main(int ac,char **av){
	ARGS		ARGS;
	MSG		MSG;
	D4LOAD		D4LD;
	D4SAVE		D4SV;
	int		span,width,height;
	int		retcode,errflag;
	d4file		d4header;
	d4pac		d4p;
	char		*fname;

	if ( ac < 2 ){
		MSG.usage();
		exit(0);
	}
	if ( 0 != ARGS.scanargv(ac,av) ){
		MSG.switcherr();
		exit(1);
	}

	ARGS.infile_rewind();
	retcode = D4LD.fileopen( fname = ARGS.infile_next() );
	switch ( retcode ){
	  case 0:	break;
	  case -3:	MSG.nonsupported(fname);  exit(2);
	  case -4:	MSG.memoryover();  exit(3);
	  default:	MSG.readerr(fname);  exit(2);
	}

	D4LD.getupper(&d4p);
	span   = d4p.d4_data[2];
	width  = d4p.d4_data[0] +1;
	height = d4p.d4_data[1] +1;

	D4LD.fileclose();

	d4header.d4_llimit[0] = 0;		d4header.d4_hlimit[0] = width -1;
	d4header.d4_llimit[1] = 0;		d4header.d4_hlimit[1] = height -1;
	d4header.d4_llimit[2] = span;		d4header.d4_hlimit[2] = span;
	d4header.d4_llimit[3] = 1;		d4header.d4_hlimit[3] = height;
	if ( 0 != D4SV.filecreat(ARGS.outfile(),&d4header) ){
		MSG.createrr(ARGS.outfile());
		exit(2);
	}

	ARGS.infile_rewind();

	errflag = 0;
	while ( 0 != (fname = ARGS.infile_next()) ){
		retcode = D4LD.fileopen( fname );
		switch ( retcode ){
		  case 0:	break;
		  case -3:	MSG.nonsupported(fname);  errflag = -1;  break;
		  case -4:	MSG.memoryover();  errflag = -1;  break;
		  default:	MSG.readerr(fname);  errflag = -1;  break;
		}
		if ( errflag )  break;

		D4LD.rewind();
		while ( 0 == D4LD.getdata(&d4p) ){
			if ( 0 != D4SV.putdata(&d4p) ){
				MSG.writeerr(ARGS.outfile());
				errflag = -1;
				break;
			}
		}
		D4LD.fileclose();
	}

	if ( 0 != D4SV.fileclose() )  errflag = -1;
	if ( errflag ){
		unlink(ARGS.outfile());
	}

	return(0);
}
