/*--------------------------------------------------------------
	Primitive Rectangle filter   bxfil  v1.4
		Written  by H.Goto , July 1994
		Modified by H.Goto , Apr. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#define		MAXGLBOXES	16384

#define		BXSTAT_LINKED	0x80


#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>

#include	"ufilep.h"
#include	"DataFile.h"


struct boxlist{
	int	stat;
	int	link;
	int	xoff;
	int	yoff;
	int	width;
	int	height;
	long	pixsum;
};



/*------------------------------------------------------
	Box list buffer class
------------------------------------------------------*/

class BXLST {
    private:
	boxlist		*bxlst;
	int		*bxlab;
	int		boxes;
	int		maxboxes;
	int		getcounter;
    public:
	int		alloc(int boxes);
	void		dealloc(void);
	int		bxadd(int x0,int y0,int width,int height);
	int		getboxes(void){  return boxes; }
	void		rewind(void){  getcounter = 0; }
	int		getnext (boxlist *bxp);
	void		c2cat(int span,int columns);
			BXLST(void);
			~BXLST(void);
};


BXLST :: BXLST (void){
	bxlst = 0;
	bxlab = 0;
	maxboxes = 0;
}


BXLST :: ~BXLST (void){
	dealloc();
}


void BXLST :: dealloc(void){
	if ( bxlst != 0 )  delete []bxlst;
	if ( bxlab != 0 )  delete []bxlab;
	maxboxes = 0;
}


int BXLST :: alloc(int boxes){
	dealloc();
	if ( (bxlst = new boxlist[boxes]) == 0 )  return(-1);
	if ( (bxlab = new int    [boxes]) == 0 )  return(-1);
	maxboxes = boxes;
	(int)BXLST::boxes = 0;
	return(0);
}


int BXLST :: bxadd(int x0,int y0,int width,int height){
	if ( boxes >= maxboxes )  return(-1);
	bxlst[boxes].stat = 0;
	bxlst[boxes].xoff = x0;
	bxlst[boxes].yoff = y0;
	bxlst[boxes].width = width;
	bxlst[boxes].height = height;
	++boxes;
	return(0);
}


int BXLST :: getnext(boxlist *bxp){
	for ( ; getcounter<boxes ; getcounter++ ){
		if ( (bxlst[getcounter].stat & BXSTAT_LINKED) == 0 ){
			*bxp = bxlst[ getcounter++ ];
			return(0);
		}
	}
	return(-1);
}


void BXLST :: c2cat(int span,int columns){
	int	ix,ix2,xc,xc2;
	int	y01,y11,y02,y12;
	int	bxs,ix3,ix4,ix40,ix5,ix50;
	for ( ix3=0 ; ix3<columns ; ix3++ ){
		for ( ix=bxs=0 ; ix<boxes ; ix++ ){
//			if ( (bxlst[ix].stat & BXSTAT_LINKED) != 0 )  continue;
			if ( (bxlst[ix].xoff /span /2) != ix3 )  continue;
			bxlab[ bxs++ ] = ix;
		}
		for ( ix4=0 ; ix4<(bxs-1) ; ix4++ ){
			ix = bxlab[ ix40 = ix4 ];
			while ( (bxlst[ix].stat & BXSTAT_LINKED) != 0 ){
				ix40 = bxlst[ix].link;
				ix = bxlab[ ix40 ];
			}
			xc = bxlst[ix].xoff / span;
			xc2 = (xc ^ 0x01) * span;
			xc = (xc & (~0x01)) * span;
			y01 = bxlst[ix].yoff;
			y11 = y01 + bxlst[ix].height -1;
			for ( ix5=ix4+1 ; ix5<bxs ; ix5++ ){
				ix2 = bxlab[ ix50 = ix5 ];
				while ( (bxlst[ix2].stat & BXSTAT_LINKED) != 0 ){
					ix50 = bxlst[ix2].link;
					ix2 = bxlab[ ix50 ];
				}
				if ( ix50 == ix40 )  continue;
				y02 = bxlst[ix2].yoff;
				y12 = y02 + bxlst[ix2].height -1;
				if ( y11 < y02 )  continue;
				if ( y12 < y01 )  continue;
				if ( y01 > y02 )  y01 = y02;
				if ( y11 < y12 )  y11 = y12;
				bxlst[ix].xoff = xc;
				bxlst[ix].width = span *2;
				bxlst[ix].yoff = y01;
				bxlst[ix].height = y11 - y01 +1;
				bxlst[ix2].stat |= BXSTAT_LINKED;
				bxlst[ix2].link = ix40;
			}
		}
	}
	for ( ix=0 ; ix<boxes ; ix++ ){
		if ( (bxlst[ix].stat & BXSTAT_LINKED) != 0 )  continue;
		if ( bxlst[ix].width == (span * 2) )  continue;
		bxlst[ix].width = (span * 2);
		xc = (bxlst[ix].xoff / span) & ~0x01;
		bxlst[ix].xoff = xc * span;
	}
}




/*------------------------------------------------------
	Message class
------------------------------------------------------*/

class MSG {
    public:
	void		usage(void);
	void		switcherr(void);
	void		readerr(char *name);
	void		writeerr(char *name);
	void		createrr(char *name);
	void		nonsupported(char *name);
	void		memoryover(void);
	void		toomanyboxes(void);
};
    	
    	
void MSG :: usage(void){		// Display Usage
	fputs("bxfil  v1.4,  Copyright (C) 1994-1997 Hideaki Goto \n",stderr);
	fputs("Usage : bxfil  [-options] in_file|stdin outfile \n",stderr);
	fputs("		-double  : double the resolution \n",stderr);
	fputs("		-halve   : halve the resolution  \n",stderr);
	fputs("		-d2      : divide into 2 columns \n",stderr);
	fputs("		-c2      : combine 2 columns \n",stderr);
//	fputs("		-quiet   : suppress messages \n",stderr);
}


void MSG :: switcherr(void){
	fputs("Switch error.\n",stderr);
}


void MSG :: readerr(char *name){
	fprintf(stderr,"Can't read %s.\n",name);
}


void MSG :: writeerr(char *name){
	fprintf(stderr,"Can't write %s.\n",name);
}


void MSG :: createrr(char *name){
	fprintf(stderr,"Can't create %s.\n",name);
}


void MSG :: nonsupported(char *name){
	fprintf(stderr,"Non-supported file %s.\n",name);
}


void MSG :: memoryover(void){
	fputs("Not enough memory.\n",stderr);
}


void MSG :: toomanyboxes(void){
	fputs("Too many boxes.\n",stderr);
}



/*------------------------------------------------------
	Argument / Option class
------------------------------------------------------*/

class ARGS {
    private:
	char		*ifname;
	char		*ofname;
    	int		sw_verbose;
	int		sw_double;
	int		sw_span;
	int		sw_d2;
	int		sw_c2;
	int		sw_halve;
    public:
	int		scanargv(int ac,char **av);
	int		SWverbose(void) {  return sw_verbose; }
	int		SWdouble(void) {  return sw_double; }
	int		SWspan(void) {  return sw_span; }
	int		SWd2(void) {  return sw_d2; }
	int		SWc2(void) {  return sw_c2; }
	int		SWhalve(void) {  return sw_halve; }
	char		*infile(void);
	char		*outfile(void);
			ARGS(void);
};


ARGS :: ARGS(void){
	ifname = ofname = 0;
	sw_verbose = -1;
	sw_double = 0;
	sw_span = 8;
	sw_d2 = 0;
	sw_c2 = 0;
	sw_halve = 0;
}


int ARGS :: scanargv(int ac,char **av){
	int	i,err,skip;
	char	*sw;
	err = 0;
	skip = 0;
	for( i = 1 ; i < ac ; ++i ){
		if( skip ){  skip = 0;  continue;  }
		if( *( sw = av[i] ) == '-' ){
			++sw;
			if( 0 == strcmp(sw,"span") ){
				if(i<ac){
					sw_span = atoi(av[i+1]);
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"halve") ){
				sw_halve = -1;
				continue;
			}
			if( 0 == strcmp(sw,"d2") ){
				sw_d2 = -1;
				continue;
			}
			if( 0 == strcmp(sw,"c2") ){
				sw_c2 = -1;
				continue;
			}
			if( 0 == strcmp(sw,"double") ){
				sw_double = -1;
				continue;
			}
			if( 0 == strcmp(sw,"quiet") ){
				sw_verbose = 0;
				continue;
			}
			if( 0 == strcmp(sw,"verbose") ){
				sw_verbose = -1;
				continue;
			}
			err = -1;  break;
		}
		else{	if ( ifname == 0 )  ifname = av[i];
			else{	ofname = av[i];  break;  }
		}
	}
	return(err);
}


char * ARGS :: infile(void){
	if ( ifname == 0 )  return( "" );
	return(ifname);
}


char * ARGS :: outfile(void){
	if ( ofname == 0 )  return( "/dev/null" );
	return(ofname);
}




/*------------------------------------------------------
	Main routine
------------------------------------------------------*/


int main(int ac,char **av){
	ARGS		ARGS;
	MSG		MSG;
	BXLST		BXLST;
	D4LOAD		D4LD;
	D4SAVE		D4SV;
	int		span,span0,width,height;
	int		retcode,errflag;
	d4file		d4header;
	D8Rect		d8pr;
	int		dims;
	int		columns;
	boxlist		bxp;

	if ( ac < 2 ){
		MSG.usage();
		exit(0);
	}
	if ( 0 != ARGS.scanargv(ac,av) ){
		MSG.switcherr();
		exit(1);
	}

	if ( 0 != BXLST.alloc(MAXGLBOXES) ){
		MSG.memoryover();
		exit(3);
	}

	retcode = D4LD.fileopen(ARGS.infile());
	switch ( retcode ){
	  case 0:	break;
	  case -3:	MSG.nonsupported(ARGS.infile());  exit(2);
	  case -4:	MSG.memoryover();  exit(3);
	  default:	MSG.readerr(ARGS.infile());  exit(2);
	}

//	errflag = 0;

	D4LD.getupper((d4pac *)(&d8pr));
//	fprintf(stderr,"Upper limits : %d\t%d\t%d\t%d\n",d8pr.x0,d8pr.y0,d8pr.w,d8pr.h);
	span = span0 = d8pr.w;
	width  = d8pr.x0 +1;
	height = d8pr.y0 +1;
	if ( ARGS.SWc2() ){
		span *= 2;
	}
	columns = (width + span -1) / span;
	width  = columns * span;
	d8pr.pixsum = 0;
	d8pr.d0 = d8pr.d1 = 0;

	if ( ARGS.SWhalve() ){
		width  = (width  +1) /2;
		height = (height +1) /2;
		span /= 2;
	}

	d4header.d4_llimit[0] = 0;		d4header.d4_hlimit[0] = width -1;
	d4header.d4_llimit[1] = 0;		d4header.d4_hlimit[1] = height -1;
	d4header.d4_llimit[2] = span;		d4header.d4_hlimit[2] = span;
	d4header.d4_llimit[3] = 1;		d4header.d4_hlimit[3] = height;
	if ( 0 != D4SV.filecreat(ARGS.outfile(),&d4header) ){
		MSG.createrr(ARGS.outfile());
		exit(2);
	}

	if ( ! ARGS.SWc2() ){
		D4LD.rewind();
		errflag = 0;
		while ( 0 == D4LD.getdata((d4pac *)(&d8pr)) ){
			if ( ARGS.SWdouble() ){
				d8pr.x0 *= 2;	d8pr.y0 *= 2;
				d8pr.w  *= 2;	d8pr.h  *= 2;
//				d8pr.pixsum *= 4;
			}
			if ( ARGS.SWhalve() ){
				d8pr.x0 = (d8pr.x0 + 1) /2;
				d8pr.y0 = (d8pr.y0 + 1) /2;
				d8pr.w  = (d8pr.w  + 1) /2;
				d8pr.h  = (d8pr.h  + 1) /2;
//				d8pr.pixsum = (d8pr.pixsum + 2) /4;
			}
			if ( ARGS.SWd2() && (d8pr.w != 0) ){
				d8pr.w = (d8pr.w +1) /2;
				if ( 0 != D4SV.putdata((d4pac *)(&d8pr)) )  errflag = -1;
				d8pr.x0 += d8pr.w;
			}
			if ( 0 != D4SV.putdata((d4pac *)(&d8pr)) )  errflag = -1;
			if ( errflag ){
				MSG.writeerr(ARGS.outfile());
				break;
			}
		}
	}
	else{	D4LD.rewind();
		errflag = 0;
		while ( 0 == D4LD.getdata((d4pac *)(&d8pr)) ){
			if ( d8pr.w == 0 )  continue;
			if ( ARGS.SWdouble() ){
				d8pr.x0 *= 2;	d8pr.y0 *= 2;
				d8pr.w  *= 2;	d8pr.h  *= 2;
//				d8pr.pixsum *= 4;
			}
			if ( BXLST.bxadd(d8pr.x0,d8pr.y0,d8pr.w,d8pr.h) != 0 ){
				MSG.toomanyboxes();
				errflag = -1;
				break;
			}
		}
		if ( ! errflag ){
			BXLST.c2cat(span0,columns);
			BXLST.rewind();
			while ( 0 == BXLST.getnext(&bxp) ){
				d8pr.x0 = bxp.xoff;
				d8pr.y0 = bxp.yoff;
				d8pr.w  = bxp.width;
				d8pr.h  = bxp.height;
				if ( ARGS.SWhalve() ){
					d8pr.x0 = (d8pr.x0 + 1) /2;
					d8pr.y0 = (d8pr.y0 + 1) /2;
					d8pr.w  = (d8pr.w  + 1) /2;
					d8pr.h  = (d8pr.h  + 1) /2;
//					d8pr.pixsum = (d8pr.pixsum + 2) /4;
				}
				if ( ARGS.SWd2() ){
					d8pr.w = (d8pr.w +1) /2;
					if ( 0 != D4SV.putdata((d4pac *)(&d8pr)) )  errflag = -1;
					d8pr.x0 += d8pr.w;
				}
				if ( 0 != D4SV.putdata((d4pac *)(&d8pr)) )  errflag = -1;
				if ( errflag ){
					MSG.writeerr(ARGS.outfile());
					break;
				}
			}
		}
	}

	if ( 0 != D4SV.fileclose() )  errflag = -1;
	if ( errflag ){
		unlink(ARGS.outfile());
	}

	D4LD.fileclose();
	return(0);
}
