/*------------------------------------------------------
	Library for Document Analysis System
	(Extended Math. class)
		Written  by H.Goto, Nov. 1994
		Modified by H.Goto, Feb. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<stdlib.h>

#include	"Exmath.h"


int ExMATH :: lsinit(int ar,int br,int alimit){
	(int)ExMATH::ar = ar;
	(int)ExMATH::br = br;
	(int)ExMATH::alimit = abs(alimit);
	xsum = ysum = xxsum = xysum = 0;
	counts = 0;
	return (0);
}


int ExMATH :: lsputdata(int x,int y){
	xsum += x;
	ysum += y;
	xxsum += (x * x);
	xysum += (x * y);
	counts += 1;
	return (0);
}


int ExMATH :: lsputdata(int x,int y,int wt){	/* Weighted version */
	xsum += (x * wt);
	ysum += (y * wt);
	xxsum += (x * x * wt * wt);
	xysum += (x * y * wt * wt);
	counts += wt;
	return (0);
}


int ExMATH :: lsgetA(void){
	int	ax,a;
	float	axf;
	ax = (counts * xxsum) - (xsum * xsum);
	if ( ax == 0 )  return (0);
	if ( ar <= 256 ){
		a = ar * ((counts * xysum) - (xsum * ysum)) / ax;
	}
	else{
		axf = ((float)counts * (float)xxsum) \
				- ((float)xsum * (float)xsum);
		a = (int)((float)ar * (((float)counts * (float)xysum) \
			- ((float)xsum * (float)ysum)) / axf);
	}
	if ( a < -alimit )  a = -alimit;
	if ( a >  alimit )  a =  alimit;
	return (a);
}


#if 0
int ExMATH :: lsgetB(void){
	int	ax;
	ax = (counts * xxsum) - (xsum * xsum);
	if ( ax == 0 )  return (br * ysum / counts);
	return ( br * ((xxsum * ysum) - (xysum * xsum)) / ax );
}
#else
int ExMATH :: lsgetB(void){
	int	a;
	a = (xsum * lsgetA()) /ar;
	return ( (ysum - a) / counts );
}
#endif


int ExMATH :: lsgetB2(int a){
	int	aa;
	aa = (xsum * a) /ar;
	return ( (ysum - aa) / counts );
}


int ExMATH :: lncross(int x0,int y0,int x1,int y1,int x2,int y2,int x3,int y3){
	int	D1,D2,D3,D4;
	D1 = ( (x1 - x0) * (y3 - y2) - (y1 - y0) * (x3 - x2) );
	D3 = D1;
	D2 = ( (x1 - x0) * (y0 - y2) - (y1 - y0) * (x0 - x2) );
	D4 = ( (x3 - x2) * (y0 - y2) - (y3 - y2) * (x0 - x2) );
	if ( D1 == 0 )  return(0);
#if 0
	if ( (D1 * D2) <= 0 )  return(0);
	if ( abs(D1) <= abs(D2) )  return(0);
	if ( (D3 * D4) <= 0 )  return(0);
	if ( abs(D3) <= abs(D4) )  return(0);
#else
	if ( (D1 * D2) < 0 )  return(0);
	if ( abs(D1) < abs(D2) )  return(0);
	if ( (D3 * D4) < 0 )  return(0);
	if ( abs(D3) < abs(D4) )  return(0);
#endif
	return(-1);
}


#if 0
int ExMATH :: lncross(int x0,int y0,int x1,int y1,int x2,int y2,int x3,int y3){
	int	D1,D2,D3,D4;
	D1 = ( (x1 - x0) * (y3 - y2) - (y1 - y0) * (x3 - x2) );
	D2 = ( (x3 - x2) * (y0 - y2) - (y3 - y2) * (x0 - x2) );
	D3 = ( (x3 - x2) * (y1 - y0) - (y3 - y2) * (x1 - x0) );
	D4 = ( (y1 - y0) * (x0 - x2) - (x1 - x0) * (y0 - y2) );
	if ( (D1 * D3) == 0 )  return(0);
#if 0
	if ( (D1 * D2) <= 0 )  return(0);
	if ( abs(D1) <= abs(D2) )  return(0);
	if ( (D3 * D4) <= 0 )  return(0);
	if ( abs(D3) <= abs(D4) )  return(0);
#else
	if ( (D1 * D2) < 0 )  return(0);
	if ( abs(D1) < abs(D2) )  return(0);
	if ( (D3 * D4) < 0 )  return(0);
	if ( abs(D3) < abs(D4) )  return(0);
#endif
	return(-1);
}
#endif

