/*------------------------------------------------------
	Library for Document Analysis System
	(Text-line object file I/O class)
		Written  by H.Goto, Aug. 1996
		Modified by H.Goto, Feb. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1996-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<string.h>

#include	<ufilep.h>
#include	"TLfile.h"


void TLFILE :: getsize(int *width,int *height){
	*width = size.d4_hlimit[0];
	*height = size.d4_hlimit[1];
}


void TLFILE :: setsize(int lx,int ly,int lwidth,int lheight,\
			int hx,int hy,int hwidth,int hheight){
	size.d4_llimit[0] = lx;
	size.d4_llimit[1] = ly;
	size.d4_llimit[2] = lwidth;
	size.d4_llimit[3] = lheight;
	size.d4_hlimit[0] = hx;
	size.d4_hlimit[1] = hy;
	size.d4_hlimit[2] = hwidth;
	size.d4_hlimit[3] = hheight;
}


int TLFILE :: read_textline(TLOBJ **tl){
	TLOBJ	*tlobj;
	d4pac	d4p;
	int	retcode;
	TLrect	rect;
	tlobj = 0;
	retcode = 0;
	while(1){
		if ( getdata(&d4p) ){
			retcode = -1;  break;	/* EOF or Read Error */
		}

		if ( d4p.d4_data[0] == 0 && d4p.d4_data[1] == 0 \
		  && d4p.d4_data[2] == 0 && d4p.d4_data[3] == 0 )  break;

		if ( d4p.d4_data[0] == -1 && d4p.d4_data[1] == -1 \
		  && d4p.d4_data[2] == 0 && d4p.d4_data[3] == 0 ){
			retcode = 1;		/* End of Block */
			break;
		}

		if ( tlobj == 0 ){
			if ( 0 == (tlobj = new TLOBJ) ){
				retcode = -2;	/* Memory error */
				break;
			}
			tlobj->stat |= TLSTAT_ACTIVE;
		}
		rect.x = d4p.d4_data[0];
		rect.y = d4p.d4_data[1];
		rect.w = d4p.d4_data[2];
		rect.h = d4p.d4_data[3];
		if ( 0 > tlobj->addrects(1,&rect) ){
			retcode = -2;		/* Memory error */
			delete tlobj;
			tlobj = 0;
			break;
		}
	}
	*tl = tlobj;
	return(retcode);
}


int TLFILE :: load(char *fname,TLGroup *TLGRP){
	TLOBJ	*tlobj;
	d4pac	d4p;
	int	errflag;
	int	rc;
	TLrect	rect;
	if ( TLGRP->deletealllines() )  return(-1);
	tlobj = 0;
	if ( fileopen(fname) )  return(-1);
	getlower((d4pac *)size.d4_llimit);
	getupper((d4pac *)size.d4_hlimit);
	errflag = 0;
	while(1){
		rc = read_textline(&tlobj);
		if ( rc < -1 ){  errflag = 1;  break; }
		if ( tlobj != 0 ){
			if ( 0 > TLGRP->addline(tlobj) ){
				errflag = 1;
				break;
			}
			tlobj = 0;
		}
		if ( rc == -1 )  break;
	}
	if ( !errflag && tlobj != 0 ){
		if ( 0 > TLGRP->addline(tlobj) )  errflag = 1;
	}
	D4LOAD::fileclose();
	return(errflag);
}


int TLFILE :: load_blocks(char *fname,TLGroup ***TLGlist){
	TLOBJ	*tlobj;
	d4pac	d4p;
	int	errflag;
	int	rc;
	TLrect	rect;
	TLGroup	**TLGs = 0;
	typedef struct strc_TLGplist{
		TLGroup			*TLG;
		struct strc_TLGplist	*next;
	} TLGplist;
	TLGplist	*TLGroot,*TLGcurrent;
	int	gcounts,i;
	if ( 0 == (TLGroot = new TLGplist) )  return(-1);
	TLGroot->TLG = 0;
	TLGroot->next = 0;
	TLGcurrent = TLGroot;

	tlobj = 0;
	if ( fileopen(fname) )  return(-1);
	getlower((d4pac *)size.d4_llimit);
	getupper((d4pac *)size.d4_hlimit);
	errflag = 0;
	while(1){
		rc = read_textline(&tlobj);
		if ( rc < -1 ){  errflag = -1;  break; }
		if ( tlobj != 0 ){
			if ( TLGcurrent->TLG == 0 ){
				if ( 0 == (TLGcurrent->TLG = new TLGroup) ){
					errflag = -1;
					break;
				}
			}
			if ( 0 > TLGcurrent->TLG->addline(tlobj) ){
				errflag = -1;
				break;
			}
			tlobj = 0;
		}
		if ( rc == -1 )  break;		/* EOF */
		if ( rc == 1 ){
			if (  0 == (TLGcurrent->next = new TLGplist) ){
				errflag = -1;
				break;
			}
			TLGcurrent = TLGcurrent->next;
			TLGcurrent->TLG = 0;
			TLGcurrent->next = 0;
		}
	}
	if ( !errflag && tlobj != 0 ){
		if ( 0 > TLGcurrent->TLG->addline(tlobj) )  errflag = -1;
	}
	D4LOAD::fileclose();
	if ( ! errflag ){
		for ( gcounts = 0, TLGcurrent = TLGroot ; ; ){
			if ( TLGcurrent->TLG == 0 )  break;
			gcounts++;
			if ( 0 == (TLGcurrent = TLGcurrent->next) )  break;
		}
		if ( gcounts > 0 ){
			if ( 0 == (TLGs = new TLGroup*[gcounts]) ){
				gcounts = -1;
			}
			else{	for ( i=0, TLGcurrent = TLGroot ; i<gcounts ; i++ ){
					TLGs[i] = TLGcurrent->TLG;
					TLGcurrent = TLGcurrent->next;
				}
			}
		}
		*TLGlist = TLGs;
		errflag = gcounts;
	}
	while(1){
		TLGcurrent = TLGroot;
		TLGroot = TLGcurrent->next;
		delete TLGcurrent;
		if ( 0 == TLGroot )  break;
	}
	return(errflag);
}


int TLFILE :: save(char *fname,TLGroup *TLGRP,int tlmaskmode){
	TLOBJ	*tlobj;
	TLrect	*rect;
	d4pac	d4p;
	int	errflag;
	int	line,i;
	if ( 0 > filecreat(fname,&size) )  return(-1);
	errflag = 0;
	if ( TLGRP != 0 ){
		for ( line=0 ; line < TLGRP->lines && !errflag ; line++ ){
			if ( 0 == (tlobj = TLGRP->tllist[line]) )  continue;
			if ( tlmaskmode != 0 ){
				if ( ! (tlobj->stat & TLSTAT_ACTIVE) )  continue;
			}
			for ( i=0 ; i < tlobj->rects ; i++ ){
				rect = tlobj->rectlist + i;
				d4p.d4_data[0] = rect->x;
				d4p.d4_data[1] = rect->y;
				d4p.d4_data[2] = rect->w;
				d4p.d4_data[3] = rect->h;
				if ( 0 > putdata(&d4p) ){
					errflag = 1;
					break;
				}
			}
			d4p.d4_data[0] = d4p.d4_data[1] = 0;
			d4p.d4_data[2] = d4p.d4_data[3] = 0;
			if ( 0 > putdata(&d4p) ){
				errflag = 1;
				break;
			}
		}
	}
	if ( 0 > D4SAVE::fileclose() )  errflag = 1;
	return(errflag);
}




