/*------------------------------------------------------
	Library for Document Analysis System
	(TLimage:  Handle Text-Line Images)
		Written  by H.Goto, May  1996
		Modified by H.Goto, Dec. 1996
		Modified by H.Goto, Feb. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1996-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

#include	"SysConst.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<math.h>

#include	"TLimage.h"

#define		ANGLERES	1024



ImgFBD * TLImageH :: createfb(int xoff,int yoff,int width, \
					int height,int xm,int ym){
	ImgFBD	*dfb;
	width  += 2 * xm;
	height += 2 * ym;
	xoff -= xm;
	if ( xoff < 0 ){  width -= xoff;  xoff = 0; }
	yoff -= ym;
	if ( yoff < 0 ){  height -= yoff;  yoff = 0; }
	if ( 0 == (dfb = new ImgFBD) )  return(0);
	dfb->xoff = xoff;
	dfb->yoff = yoff;
	dfb->width  = width;
	dfb->height = height;
	if ( 0 == (dfb->image = sip_CreateImage(width,height,8)) ){
		delete dfb;
		return(0);
	}
	sip_ClearImage(dfb->image,0xff);
	return(dfb);
}


int TLImageH :: deletefb(ImgFBD *dfb){
	sip_DestroyImage(dfb->image);
	delete dfb;
	return(0);
}


int TLImageH :: save(char *fname,ImgFBD *dfb){
	int	errflag,retcode,ix;
	PBMFILE	PBMSV;
	PBMSV.setsize(dfb->width,dfb->height,1);
	retcode = PBMSV.open(fname,"w");
	if ( retcode < 0 )  return( retcode );
	errflag = 0;
	for ( ix=0 ; ix < dfb->height ; ix++ ){
		if ( 0 != PBMSV.writeline_bilevel(-1,(void *)sip_getimgptr(dfb->image,ix),128) ){
			errflag = -1;
			break;
		}
	}
	errflag |= PBMSV.close();
	if ( errflag ){
		unlink(fname);  return(-1);
	}
	return(0);
}


int TLImageH :: blkcpy(int xg0,int yg0,int w,int h,ImgFBD *srcfbd,ImgFBD *dstfbd){
	int	x,y;
	int	sxg0,syg0;
	uchar	*sbuf,*dbuf;
	uchar	*lbuf;
	sxg0 = xg0;
	syg0 = yg0;
	if ( (sxg0 -= srcfbd->xoff) < 0 )  return(-1);
	if ( (syg0 -= srcfbd->yoff) < 0 )  return(-1);
	if ( (sxg0 + w) > srcfbd->width  )  return(-1);
	if ( (syg0 + h) > srcfbd->height )  return(-1);
	if ( (xg0 -= dstfbd->xoff) < 0 )  return(-1);
	if ( (yg0 -= dstfbd->yoff) < 0 )  return(-1);
	if ( (xg0 + w) > dstfbd->width  )  return(-1);
	if ( (yg0 + h) > dstfbd->height )  return(-1);

	if ( srcfbd->image->depth == 8 ){
		for ( y=0 ; y<h ; y++ ){
			sbuf = (uchar *)sip_getimgptr(srcfbd->image,y + syg0) + sxg0;
			dbuf = (uchar *)sip_getimgptr(dstfbd->image,y +  yg0) + xg0;
			for ( x=0 ; x<w ; x++ )  dbuf[x] = sbuf[x];
		}
	}
	else{
#if 1
		for ( y=0 ; y<h ; y++ ){
			sip_cvt1to8((char *)sip_getimgptr(srcfbd->image,y+syg0) + (sxg0/8),sxg0 & 0x07,(char *)sip_getimgptr(dstfbd->image,y+yg0) + xg0,w,0x00,0xff);
		}
#else
		if ( 0 == (lbuf = new uchar[w * 2]) )  return(-1);
		for ( y=0 ; y<h ; y++ ){
			sip_GetScanLine(srcfbd->image,lbuf + w,sxg0,y + syg0,w,1);
			sip_cvt1to8((char *)(lbuf + w),0,(char *)lbuf,w,0x00,0xff);
			sip_PutScanLine(dstfbd->image,lbuf, xg0,y +  yg0,w,8);
		}
		delete []lbuf;
#endif
	}

	return(0);
}


int TLImageH :: blkcpy(cbox *cb,ImgFBD *srcfbd,ImgFBD *dstfbd){
	return( blkcpy(cb->x0,cb->y0,cb->w,cb->h,srcfbd,dstfbd) );
}


int TLImageH :: getcbox(TLGroup *TLG,int lnn,cbox *cb){
	return( getcbox(TLG,lnn,cb,0) );
}


int TLImageH :: getcbox(TLGroup *TLG,int lnn,cbox *cb,TLrect *rectlist){
	int	ix,x0,x1,y0,y1;
	TLOBJ	*TL;
	TLrect	*rlst;
	if ( lnn >= TLG->lines )  return(-1);
	if ( 0 == (TL = TLG->tllist[lnn]) )  return(-1);
	if ( (TL->stat & TLSTAT_ACTIVE) == 0 )  return(0);
	if ( TL->rects == 0 )  return(-1);
	x0 = y0 = HIGHEST;
	x1 = y1 = 0;
	if ( rectlist == 0 )  rlst = TL->rectlist;  else rlst = rectlist;
	for ( ix=0 ; ix < TL->rects ; ix++ ){
		if ( x0 > rlst[ix].x )
			x0 = rlst[ix].x;
		if ( x1 < rlst[ix].x + rlst[ix].w )
			x1 = rlst[ix].x + rlst[ix].w;
		if ( y0 > rlst[ix].y )
			y0 = rlst[ix].y;
		if ( y1 < rlst[ix].y + rlst[ix].h )
			y1 = rlst[ix].y + rlst[ix].h;
	}
	cb->x0 = x0;
	cb->w  = x1 - x0;
	cb->y0 = y0;
	cb->h  = y1 - y0;
	return(0);
}


int TLImageH :: setboxfirst(TLGroup *TLG,int lnn){
	TLOBJ	*TL;
	if ( lnn >= TLG->lines )  return(-1);
	if ( 0 == (TL = TLG->tllist[lnn]) )  return(-1);
	if ( (TL->stat & TLSTAT_ACTIVE) == 0 )  return(-1);
	if ( TL->rects == 0 )  return(-1);
	getboxctr = 0;
	getboxlnn = lnn;
	curlninfo = TL;
	return(0);
}


int TLImageH :: getboxnext(cbox *cb){
	return ( getboxnext(cb,0) );
}


int TLImageH :: getboxnext(cbox *cb,TLrect *rectlist){
	TLrect	*rlst;
	if ( 0 != rectlist )  rlst = rectlist;  else rlst = curlninfo->rectlist;
	for ( ; getboxctr < curlninfo->rects ; getboxctr++ ){
		cb->x0 = rlst[getboxctr].x;
		cb->y0 = rlst[getboxctr].y;
		cb->w  = rlst[getboxctr].w;
		cb->h  = rlst[getboxctr].h;
		if ( cb->h <= 0 )  continue;
		++getboxctr;
		return(0);
	}
	return(-1);
}


void TLImageH :: lsfitAB(TLrect *rectlist,int off,int rects,int xoff,int *a,int *b){
	int	i;
	int	x,y,b0;
	lsinit(ANGLERES,1,ANGLERES);
	rectlist += off;
	b0 = (int)rectlist[0].y;
	for ( i=0 ; i < rects ; i++ ){
		x = (int)(rectlist[i].x - xoff + rectlist[i].w / 2);
		y = (int)(rectlist[i].y + rectlist[i].h / 2);
		_lsputdata(x,y - b0);
	}
	*a = lsgetA();
	*b = lsgetB() + b0;
}


					/* Warning: For PR-list ONLY. */
void TLImageH :: getuloff(TLrect *rectlist,int off,int len,int cwid, \
			int a,int b,int *uoff,int *loff){
	int	i;
	short	u,l,c,bot;
	u = l = 0;
	rectlist += off;
	for ( i=0 ; i < len ; i++ ){
		c = (short)((a * i * cwid) / ANGLERES + b);
		if ( u < c - rectlist[i].y )
				u = c - rectlist[i].y;
		bot = rectlist[i].y + rectlist[i].h - 1;
		if ( l < bot - c )
				l = bot - c;
	}
	*uoff = u;
	*loff = l;
}


					/* Warning: For PR-list ONLY. */
void TLImageH :: setcovprs(TLrect *rectlist,TLrect *rectlist0,int off, \
	 int len,int cwid,int height,int a,int b,int uoff,int loff,int mode){
	int	i,xc;
	int	center;
	xc = (len -1) /2;
	rectlist  += off;
	rectlist0 += off;
	for ( i=0 ; i < len ; i++ ){
		rectlist[i] = rectlist0[i];
		switch ( mode ){
		  case 0:	if ( i != xc )  continue;
				break;
		  case 1:	if ( i > xc )  continue;
				break;
		  case 2:	if ( i < xc )  continue;
				break;
		}
		center = (a * i * cwid) / ANGLERES + b;
		if ( 0 > (int)(rectlist[i].y = center - uoff) )  rectlist[i].y = 0;
		rectlist[i].h = uoff + loff +1;
		if ( height < (int)(rectlist[i].y + rectlist[i].h) )
			rectlist[i].h = height - rectlist[i].y;
	}
}


			/* Get skew of text line */
double TLImageH :: getskew(TLGroup *TLG,int lnn){
	TLOBJ	*TL;
	cbox	lncbox;
	int	a,b;
	if ( lnn >= TLG->lines )  return(0);
	if ( 0 == (TL = TLG->tllist[lnn]) )  return(0);
	if ( 0 != getcbox(TLG,lnn,&lncbox) )  return(0);
	lsfitAB(TL->rectlist,0,TL->rects,lncbox.x0,&a,&b);
	return( atan((double)a / ANGLERES) );
}


			 /* Get line image in Raw-mode */
ImgFBD *TLImageH :: getimage_raw(ImgFBD *docimg,TLGroup *TLG,int lnn,int mul){
	cbox	lncbox;
	ImgFBD	*strimg;
	if ( 0 != getcbox(TLG,lnn,&lncbox) )  return(0);
	lncbox.x0 *= mul;	lncbox.y0 *= mul;
	lncbox.w  *= mul;	lncbox.h  *= mul;
	if ( 0 == (strimg = createfb(lncbox.x0,lncbox.y0, \
			lncbox.w,lncbox.h,0,0)) )  return(0);

	if ( 0 == setboxfirst(TLG,lnn) ){
		while ( 0 == getboxnext(&lncbox) ){
			lncbox.x0 *= mul;	lncbox.y0 *= mul;
			lncbox.w  *= mul;	lncbox.h  *= mul;
			blkcpy(lncbox.x0,lncbox.y0,lncbox.w,lncbox.h,docimg,strimg);
		}
	}

	return(strimg);
}


			/* Get line image in Rectangular-mode */
ImgFBD *TLImageH :: getimage_rect(ImgFBD *docimg,TLGroup *TLG,int lnn,int mul){
	cbox	lncbox;
	int	x0;
	ImgFBD	*strimg;
	if ( 0 != getcbox(TLG,lnn,&lncbox) )  return(0);
	lncbox.x0 *= mul;	lncbox.y0 *= mul;
	lncbox.w  *= mul;	lncbox.h  *= mul;
	if ( 0 == (strimg = createfb(lncbox.x0,lncbox.y0, \
				lncbox.w,lncbox.h,0,0)) )  return(0);

	blkcpy(lncbox.x0,lncbox.y0,lncbox.w,lncbox.h,docimg,strimg);

	return(strimg);
}


			/* Get line image in Inclined-rectangular-mode */
			/* Warning: For PR-list ONLY. */
			/*          PR-list must be interpolated. */
			/*          PR-list must be tlproc_columnpack()ed. */
ImgFBD *TLImageH :: getimage_irect(ImgFBD *docimg,TLGroup *TLG, \
					int lnn,int cwid,int mul){
	cbox	lncbox;
	int	x0;
	ImgFBD	*strimg;
	TLrect	*prlist;
	TLOBJ	*TL;
	int	a,b;
	int	uoff,loff;
	if ( lnn >= TLG->lines )  return(0);
	if ( 0 == (TL = TLG->tllist[lnn]) )  return(0);
	if ( 0 != getcbox(TLG,lnn,&lncbox) )  return(0);
	x0 = lncbox.x0;
	lncbox.x0 *= mul;	lncbox.y0 *= mul;
	lncbox.w  *= mul;	lncbox.h  *= mul;

	if ( 0 == (prlist = new TLrect[TL->rects]) )  return(0);

	lsfitAB(TL->rectlist,0,TL->rects,x0,&a,&b);
	getuloff(TL->rectlist,0,TL->rects,cwid,a,b,&uoff,&loff);

	setcovprs(prlist,TL->rectlist,0,TL->rects,cwid, \
			docimg->height,a,b,uoff,loff,3);

	if ( 0 != getcbox(TLG,lnn,&lncbox,prlist) ){
		delete []prlist;
		return(0);
	}
	lncbox.x0 *= mul;	lncbox.y0 *= mul;
	lncbox.w  *= mul;	lncbox.h  *= mul;
	if ( 0 == (strimg = createfb(lncbox.x0,lncbox.y0,lncbox.w,lncbox.h,0,0)) ){
		delete []prlist;
		return(0);
	}

	if ( 0 == setboxfirst(TLG,lnn) ){
		while ( 0 == getboxnext(&lncbox,prlist) ){
			lncbox.x0 *= mul;	lncbox.y0 *= mul;
			lncbox.w  *= mul;	lncbox.h  *= mul;
			blkcpy(&lncbox,docimg,strimg);
		}
	}

	delete []prlist;
	return(strimg);
}


			/* Get line image in Local-rectangular-mode */
			/* Warning: For PR-list ONLY. */
			/*          PR-list must be interpolated. */
			/*          PR-list must be tlproc_columnpack()ed. */
ImgFBD *TLImageH :: getimage_lrect(ImgFBD *docimg,TLGroup *TLG, \
					int lnn,int cwid,int mul,int fmul){
	cbox	lncbox;
	int	x0;
	ImgFBD	*strimg;
	TLrect	*prlist;
	TLOBJ	*TL;
	int	a,b;
	int	uoff,loff;
	int	fspan,fpos;
	if ( lnn >= TLG->lines )  return(0);
	if ( 0 == (TL = TLG->tllist[lnn]) )  return(0);
	if ( 0 != getcbox(TLG,lnn,&lncbox) )  return(0);
	x0 = lncbox.x0;
	lncbox.x0 *= mul;	lncbox.y0 *= mul;
	lncbox.w  *= mul;	lncbox.h  *= mul;

	if ( 0 == (prlist = new TLrect[TL->rects]) )  return(0);
	fspan = (int)(((double)fmul * tlproc_getmeanheight(TL)) / cwid);
	if ( fspan > TL->rects )  fspan = TL->rects;

	for ( fpos=0 ; fpos <= (TL->rects - fspan) ; fpos++ ){
		lsfitAB(TL->rectlist,fpos,fspan,fpos * cwid + x0,&a,&b);
		getuloff(TL->rectlist,fpos,fspan,cwid,a,b,&uoff,&loff);
		setcovprs(prlist,TL->rectlist,fpos,fspan,cwid, \
				docimg->height,a,b,uoff,loff,0);
		if ( fpos == 0 ){
			setcovprs(prlist,TL->rectlist,fpos,fspan,cwid, \
					docimg->height,a,b,uoff,loff,1);
		}
		if ( fpos == TL->rects - fspan ){
			setcovprs(prlist,TL->rectlist,fpos,fspan,cwid, \
					docimg->height,a,b,uoff,loff,2);
		}
	}

	if ( 0 != getcbox(TLG,lnn,&lncbox,prlist) ){
		delete []prlist;
		return(0);
	}
	lncbox.x0 *= mul;	lncbox.y0 *= mul;
	lncbox.w  *= mul;	lncbox.h  *= mul;
	if ( 0 == (strimg = createfb(lncbox.x0,lncbox.y0,lncbox.w,lncbox.h,0,0)) ){
		delete []prlist;
		return(0);
	}

	if ( 0 == setboxfirst(TLG,lnn) ){
		while ( 0 == getboxnext(&lncbox,prlist) ){
			lncbox.x0 *= mul;	lncbox.y0 *= mul;
			lncbox.w  *= mul;	lncbox.h  *= mul;
			blkcpy(&lncbox,docimg,strimg);
		}
	}

	delete []prlist;
	return(strimg);
}


ImgFBD *TLImageH :: getimage(ImgFBD *docimg,TLGroup *TLG, \
			int lnn,int cwid,int mul,int fmul,int mode){
	switch(mode){
	    case 0:
		return( getimage_raw(docimg,TLG,lnn,mul) );
	    case 1:
		return( getimage_rect(docimg,TLG,lnn,mul) );
	    case 2:
		return( getimage_irect(docimg,TLG,lnn,cwid,mul) );
	    case 3:
		return( getimage_lrect(docimg,TLG,lnn,cwid,mul,fmul) );
	}
	return(0);
}




