/*------------------------------------------------------
	Library for Document Analysis System
	(Text-line object processing)
		Written  by H.Goto, July 1996
		Modified by H.Goto, Feb. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1996-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<string.h>

#include	"TLobj.h"

#include	"TLproc.h"


int tlproc_columnpack(TLOBJ *textline){
	int	i,rects,newrects;
	TLrect	*tmprectlist;
	int	retcode;
	int	lastcol;
	int	top,bottom,y;
	int	stat0;
	if ( textline == 0 )  return(0);
	rects = textline->rects;
	stat0 = textline->stat;
	if ( rects == 0 )  return(0);
	if ( 0 == (tmprectlist = new TLrect[rects]) )  return(-1);
	retcode = 0;
	lastcol = -1;
	newrects = 0;

	for ( i=0 ; i<rects ; i++ ){
		if ( lastcol != textline->rectlist[i].x ){
			if ( -1 != lastcol ){
				tmprectlist[newrects].x = lastcol;
				tmprectlist[newrects].y = top;
				tmprectlist[newrects].w = textline->rectlist[i-1].w;
				tmprectlist[newrects].h = bottom - top;
				++newrects;
			}
			top = textline->rectlist[i].y;
			bottom = top + textline->rectlist[i].h;
			lastcol = textline->rectlist[i].x;
		}
		else{	y = textline->rectlist[i].y;
			if ( top > y )  top = y;
			y += textline->rectlist[i].h;
			if ( bottom < y )  bottom = y;
		}
	}
	if ( -1 != lastcol ){
		tmprectlist[newrects].x = lastcol;
		tmprectlist[newrects].y = top;
		tmprectlist[newrects].w = textline->rectlist[i-1].w;
		tmprectlist[newrects].h = bottom - top;
		++newrects;
	}

	if ( textline->deleteallrects() ){
		retcode = -1;
	}
	else{	if ( 0 > textline->addrects(newrects,tmprectlist) )  retcode = -1;
	}
	textline->stat = stat0 | TLSTAT_MODIFIED;
	delete []tmprectlist;
	return(retcode);
}


double tlproc_getmeanheight(TLOBJ *TL){
	int	i;
	int	sum;
	if ( TL == 0 )  return(0);
	if ( TL->rects == 0 )  return(0);
	for ( i=0,sum=0 ; i < TL->rects ; i++ )  sum += TL->rectlist[i].h;
	return ( (double)sum / (double)TL->rects );
}


int tlproc_interp(TLOBJ *TL,int cwid){
	int	i,i2,pc,h,lt,lt0,lb,lb0;
	TLrect	*rectlist;
	int	lrect,rrect,rects;
	if ( TL == 0 )  return(0);
	if ( TL->rects == 0 )  return(0);
	lrect = 32767;  rrect = 0;
	for ( i=0 ; i < TL->rects ; i++ ){
		if ( lrect > TL->rectlist[i].x )  lrect = TL->rectlist[i].x;
		if ( rrect < TL->rectlist[i].x )  rrect = TL->rectlist[i].x;
	}
	lrect /= cwid;
	rrect /= cwid;
	rects = rrect - lrect +1;
	if ( 0 == (rectlist = new TLrect[rects]) )  return(-1);
	for ( i=0 ; i < rects ; i++ )  rectlist[i].w = 0;
	for ( i=0 ; i < TL->rects ; i++ ){
		rectlist[ (TL->rectlist[i].x / cwid) - lrect ] = TL->rectlist[i];
	}
	TL->rects = rects;
	delete []TL->rectlist;
	TL->rectlist = rectlist;
	TL->stat |= TLSTAT_MODIFIED;

	for ( i=pc=0 ; i < TL->rects ; i++ ){
		if ( TL->rectlist[i].w != 0 ){
			lt = TL->rectlist[i].y;
			lb = lt + TL->rectlist[i].h -1;
			if ( pc != 0 ){
				for ( i2=0 ; i2<pc ; i2++ ){
					TL->rectlist[i-pc+i2].x = (i -pc + i2 + lrect) * cwid;
					TL->rectlist[i-pc+i2].w = cwid;
					TL->rectlist[i-pc+i2].y = lt0 + \
						( ((lt - lt0) * (i2 + 1)) / (pc + 1) );
					h = lb0 + ( ((lb - lb0) * (i2 + 1)) / (pc + 1) );
					h += 1 - TL->rectlist[i-pc+i2].y;
					if ( h <= 0 )  h = 1;
					TL->rectlist[i-pc+i2].h = h;
				}
				pc = 0;
			}
			lt0 = lt;  lb0 = lb;
		}
		else{	++pc;
		}
	}

	return(0);
}


