/*--------------------------------------------------------------
	Text Line Extraction      lnextr2    v2.0
		Written  by H.Goto , May  1994  (v1.0)
		Modified by H.Goto , May  1996
		Modified by H.Goto , June 1996
		Modified by H.Goto , July 1997  (v2.0)
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/



//#define		_CreateAlone

//#define		_MonPeak
//#define		_MonCC
//#define		_MonCC2
//#define		_MonLNLink
//#define		_MonLNInfo
//#define		_MonPriority
//#define		_MonReject


#define		REJECT_PAR	160			/* 150 - 160 */
#define		REJECT_PAR2	250
#define		REJECT_BXS	5			/* 5 - 6 */
#define		REJECT_PAR3	200			/* 200 - 220 */

#define		CWID		16
#define		NBX		4
#define		BANDWIDTH	(NBX*CWID)
#define		ULIMIT_YSIZE0	30000
#define		LLIMIT_YSIZE0	1
#define		ULIMIT_YSIZE	100			/* = 36.14pt (200dpi) */
#define		LLIMIT_YSIZE	2			/* = 1.08pt  (200dpi) */
#define		BPRREJ_HEIGHT	5			/* lower limit for Black-PR rej. */
#define		BPRREJ_RATIO	80
#define		FILTERSPAN	7
#define		FILTER2SPAN	3
#define		PEAKTHRESHOLD	(7*3*FILTER2SPAN)	/* = 2.53pt  (200dpi) */
#define		CCSLIMIT1	2			/* lower limit par CC (create) */
#define		CCSLIMIT1A	3			/* lower limit par CC (alone) */
#define		CCSLIMIT2	3			/* lower limit par CC (attn.) */
#define		CCHLIMIT_H	70			/* = 25.3pt (200dpi) */
#define		CCHLIMIT_L	5			/* = 1.81pt (200dpi) */
#define		LNHLIMIT	26			/* = 9.40pt (200dpi) */
#define		LNLLIMIT	5			/* = 1.81pt (200dpi) */

#define		LNLINKPASS	16


#define		BOXSTEP		1024			/* Memory management config. */
#define		CCBOXES		32
#define		MAXCCS		512
#define		MAXGLCCS	4096
#define		MAXLINES	1024
#define		CCLABS_PAR_LN	128

#define		RESOLUTION_A	64
#define		RESOLUTION_AL	10
#define		RESOLUTION_B	1

#define		CCSTAT_ATTN	0x01
#define		CCSTAT_LINKED	0x02
#define		CCSTAT_KILL	0x80
#define		BXSTAT_LINKED	0x02
#define		BXSTAT_ALONE	0x08
#define		BXSTAT_CONJ	0x10
#define		BXSTAT_PASS	0x20
#define		BXSTAT_REJECT	0x80
#define		BXSTAT_CLUSTER	0x100
#define		BXSTAT_SPLITTED	0x200
#define		LNSTAT_ATTN	0x01
#define		LNSTAT_KILL	0x80


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>

#include	"stdps.h"
#include	"ufilep.h"
#include	"DataFile.h"
#include	"Exmath.h"




/*------------------------------------------------------
	Global structure definitions
------------------------------------------------------*/

struct peaklist {
	int		pos;
	int		value;
};


struct boxlist {
	int		stat;
	int		link;		// Flag for local linkage ( -1: not linked )
	int		absboxnum;	// Absolute box number
	int		boxnum;		// My box number
	int		xoff;
	int		yoff;
	int		width;
	int		height;
	long		pixsum;
};


struct ccpac {
	int		stat;
	int		link;		// Link to Line-Object
	int		xcenter;
	int		pos;		// Local position by LSM
	int		angle;		// Local angle by LSM
	int		mleft,mright;
	int		avheight;	// Real average of box height
	int		avheight2;	// Overall average of box height
	int		wsum;
	int		wsum2;
	int		boxes;
	int		columns;
	int		boxlab[CCBOXES];
};


struct lnpac {
	int		stat;
	int		x0,y0,x1,y1;
	int		leftcc,rightcc;
	int		avheight;
	int		avheight2;
	int		wsum;
	int		bsum;
	int		wsum2;
	int		csum;
	int		angle;
	int		ccs;
	int		cclab[CCLABS_PAR_LN];
};




/*------------------------------------------------------
	Box-object control class
------------------------------------------------------*/

class BXOBJ {
    private:
	int		maxboxes;
    protected:
	int		boxes;
	boxlist		*bxlst;
    public:
	boxlist		*alloc(int maxboxes);
	boxlist		*realloc(int maxboxes);
	int		dealloc(void);
	boxlist		*addbuffer(int boxes);
	boxlist		*getbuf(void){  return bxlst; }
	int		getboxes(void){  return boxes; }
	int		addnewbox(boxlist *box,int stat);
	int		addbox(boxlist *box);
	int		splitbox(int bxn,int cells);
	int		splitall(int splits);
			BXOBJ(void);
	virtual		~BXOBJ(void);
};


BXOBJ :: BXOBJ(void){
	boxes = maxboxes = 0;
	bxlst = 0;
}


BXOBJ :: ~BXOBJ(void){
	dealloc();
}


int BXOBJ :: dealloc(void){
	if ( bxlst != 0 ){
		delete	[]bxlst;
		bxlst = 0;
		maxboxes = boxes = 0;
	}
	return(0);
}


boxlist * BXOBJ :: alloc(int maxboxes){
	dealloc();
	boxes = 0;
	BXOBJ :: maxboxes = 0;
	if ( 0 == (bxlst = new boxlist[ maxboxes ]) )  return(0);
	BXOBJ :: maxboxes = maxboxes;
	return(bxlst);
}


boxlist * BXOBJ :: realloc(int maxboxes){
	int		bx;
	boxlist		*bxlst0;
	if ( BXOBJ :: maxboxes >= maxboxes )  return(bxlst);
	if ( bxlst == 0 ){
		return ( alloc(maxboxes) );
	}
	if ( 0 == (bxlst0 = new boxlist[ maxboxes ]) ){
		boxes = 0;
		BXOBJ :: maxboxes = 0;
		return(0);
	}
	for ( bx=0 ; bx<boxes ; bx++ )  bxlst0[bx] = bxlst[bx];
	BXOBJ :: maxboxes = maxboxes;
	delete []bxlst;
	return( bxlst = bxlst0 );
}


boxlist * BXOBJ :: addbuffer(int boxes){
	return( realloc( maxboxes + boxes ) );
}


int BXOBJ :: addnewbox(boxlist *box,int stat){
	if ( boxes >= maxboxes ){
		if ( 0 == addbuffer( BOXSTEP ) )  return(-1);
	}
	bxlst[boxes] = *box;
	bxlst[boxes].stat = stat;
	bxlst[boxes].link = -1;
	bxlst[boxes].boxnum = boxes;
	++boxes;
	return(boxes);
}


int BXOBJ :: addbox(boxlist *box){
	if ( boxes >= maxboxes ){
		if ( 0 == addbuffer( BOXSTEP ) )  return(-1);
	}
	bxlst[boxes] = *box;
	++boxes;
	return(boxes);
}


int BXOBJ :: splitbox(int bxn,int cells){
	int	i,y,height,yoff0;
	boxlist	box;
	if ( cells < 2 )  return(0);
	if ( bxn >= boxes )  return(0);
	if ( (bxlst[bxn].stat & BXSTAT_CLUSTER) != 0 )  return(0);
	if ( (boxes - cells) >= maxboxes )  return(-1);
	bxlst[bxn].stat = BXSTAT_SPLITTED;
	bxlst[bxn].link = -1;
	bxlst[bxn].pixsum /= cells;
	box = bxlst[bxn];
	height = box.height;
	yoff0  = box.yoff;
	for ( i=1 ; i<=cells ; i++ ){
		y = (i * height) / cells + yoff0;
		box.height = y - box.yoff;
		addnewbox(&box,BXSTAT_CLUSTER);
		box.yoff += box.height;
	}
	return(0);
}


int BXOBJ :: splitall(int splits){
	int	i;
	for ( i=0 ; i<boxes ; i++ ){
		if ( (bxlst[i].stat & BXSTAT_CONJ) == 0 )  continue;
		if ( 0 != splitbox(i,splits) )  return(-1);
	}
	return(0);
}




/*------------------------------------------------------
	Connection_Component control class
------------------------------------------------------*/

class CCOBJ : private ExMATH {
    private:
	int		setCCinfo(int bandoff,ccpac *cclist,boxlist *bxlst);
	int		mergeCC(int ccn1,int ccn2,ccpac *cclist,boxlist *bxlst);
    public:
	int		createCC(int bandoff,int boxes,ccpac *cclist, \
					boxlist *bxlst,int *rectlab);
	int		destroyCC(ccpac *cclist,boxlist *bxlst,int reason);
	int		destroyCC_ex(ccpac *cclist,boxlist *bxlst);
	int		checkCC(int bandoff,int cccounts,ccpac *cclist,boxlist *bxlst, \
							int peaks,peaklist *pks);
	int		checkCC_2(int bandoff,int cccounts,ccpac *cclist,boxlist *bxlst, \
							int peaks,peaklist *pks);
	int		rejectBox(lnpac *lnlist,int cccounts,ccpac *cclist,boxlist *bxlst0);
};


int CCOBJ :: destroyCC(ccpac *cclist,boxlist *bxlst,int reason){
	int	ix,bxn;
	for ( ix=0 ; ix<cclist->boxes ; ix++ ){
		bxn = cclist->boxlab[ix];
		bxlst[bxn].link = -1;
		bxlst[bxn].stat &= ~ BXSTAT_LINKED;
		bxlst[bxn].stat |= reason;
	}
	cclist->stat = 0;
	cclist->boxes = 0;
	return(0);
}


int CCOBJ :: destroyCC_ex(ccpac *cclist,boxlist *bxlst){
	int	reason;
	reason = 0;
	if ( cclist->boxes < CCSLIMIT1A ){
#ifdef	_CreateAlone
		reason = BXSTAT_REJECT | BXSTAT_ALONE;
#else
		reason = BXSTAT_REJECT | BXSTAT_PASS;
#endif
	}
	return ( destroyCC(cclist,bxlst,reason) );
}


int CCOBJ :: rejectBox(lnpac *lnlist,int cccounts,ccpac *cclist,boxlist *bxlst0){
	int	ix,ix2,bxn,avwid,wid,rejectn,lnn;
	int	rccs;
	int	rcc[4];
	rejectn = 0;
	for ( ix=0 ; ix<cccounts ; ix++ ){
		if ( cclist[ix].boxes == 0 )  continue;
		if ( (cclist[ix].stat & CCSTAT_LINKED ) == 0 )  continue;

		lnn = cclist[ix].link;
		avwid = (lnlist[lnn].avheight + 3*lnlist[lnn].avheight2) /4;

		if ( cclist[ix].avheight2 * 100 <= \
		     lnlist[lnn].avheight * REJECT_PAR3 )  continue;
			/* Avoid rejecting rectangles in the same text-line */

		rccs = 0;
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			bxn = cclist[ix].boxlab[ix2];
			wid = bxlst0[bxn].height;
			if ( (wid *100) <= (avwid *REJECT_PAR2) ){
				if ( cclist[ix].boxes < REJECT_BXS )  continue;
			}
			if ( (wid *100) > (avwid *REJECT_PAR) ){
				if ( rccs < 4 )  rcc[rccs++] = bxn;
			}
		}
		if ( rccs >= 3 )  continue;
			/* Limit the number of rectangles to be erased */

		for ( ix2=0 ; ix2<rccs ; ix2++ ){
			bxn = rcc[ix2];
			bxlst0[bxn].stat |= (BXSTAT_REJECT | BXSTAT_CONJ);
			++rejectn;
#ifdef _MonReject
			printf("PR %d (%d,%d) rejected.\n",bxn, \
					bxlst0[bxn].xoff,bxlst0[bxn].yoff);
#endif
		}
		if ( 0 != rccs ){
			cclist[ix].stat |= CCSTAT_KILL;
			lnlist[lnn].stat |= LNSTAT_KILL;
		}
	}
	return ( rejectn );
}


int CCOBJ :: setCCinfo(int bandoff,ccpac *cclist,boxlist *bxlst){
	int	ix,ix2,avwid,wid,pos,bxn,ml,mr;
	int	lower[NBX];
	int	higher[NBX];
	for ( ix=0 ; ix<NBX ; ix++ )  higher[ix] = -(lower[ix] = 32767);
	for ( ix=0 ; ix<(cclist->boxes) ; ix++ ){
		bxn = cclist->boxlab[ix];
		ix2 = (bxlst[bxn].xoff - bandoff) /CWID;
		pos = bxlst[bxn].yoff;
		if ( lower [ix2] > pos )  lower [ix2] = pos;
		pos += bxlst[bxn].height;
		if ( higher[ix2] < pos )  higher[ix2] = pos;
	}
	ix2 = NBX;
	avwid = 0;
	ml = -1;
	mr = 0;
	for ( ix=0 ; ix<NBX ; ix++ ){
		wid = higher[ix] - lower[ix];
		if ( wid < 0 ){ --ix2;  continue; }
		else{	avwid += wid;
			if ( ml == -1 )  ml = ix;
			mr = ix;
		}
	}
	cclist->columns = ix2;
	cclist->wsum2 = avwid;
	if ( ix2 != 0 )  avwid /= ix2;
	cclist->avheight2 = avwid;
	cclist->mleft  = ml;
	cclist->mright = NBX -1 - mr;
	return(avwid);
}


int CCOBJ :: mergeCC(int ccn1,int ccn2,ccpac *cclist,boxlist *bxlst){
	int	ix,bxs,boxes;
	int	bxn;
	bxs   = cclist[ccn1].boxes;
	boxes = cclist[ccn2].boxes;
	if ( (bxs + boxes) > CCBOXES )  return(-1);
	for ( ix=0 ; ix<boxes ; ix++ ){
		bxn = cclist[ccn2].boxlab[ix];
		cclist[ccn1].boxlab[ bxs++ ] = bxn;
		bxlst[bxn].link = ccn1;
	}
	cclist[ccn1].boxes = bxs;
	cclist[ccn2].boxes = 0;
	return(0);
}


/*------------------------------  Not included in class CCOBJ. */
static boxlist *bxlst_forcmp;

static int cmp_prpos(const void *n1,const void *n2){
	return( bxlst_forcmp[*(int *)n1].xoff - bxlst_forcmp[*(int *)n2].xoff );
}
/*-------------------------------------------------------------*/


int CCOBJ :: createCC(int bandoff,int boxes,ccpac *cclist,boxlist *bxlst,int *rectlab){
						/* Note: rectlab destructive */
	int	ix,i,j,cccounts,warn,warn2;
	int	ccn,cc0,pos,pos2,ovl,xdist;
	int	*ipr;
	boxlist	*pr1,*pr2;

	if ( rectlab == 0 ){
		if ( 0 == (ipr = new int[boxes]) )  return(0);	/* error */
		for ( ix=0 ; ix<boxes ; ix++ )  ipr[ix] = ix;
	}
	else{	ipr = rectlab;
	}
	bxlst_forcmp = bxlst;
	qsort((void *)ipr,boxes,sizeof(int),cmp_prpos);

	warn = warn2 = 0;
	cccounts = 0;
	for ( i=0 ; i<boxes ; i++ ){
		ix = ipr[i];
		pr1 = &bxlst[ix];
		if ( -1 == (cc0 = pr1->link) ){
			if ( cccounts >= MAXCCS ){
				if ( ! warn ){
					warn = -1;
					fputs("Warning: Too many CCs.\n",stderr);
				}
				break;
			}
			pr1->link = cc0 = cccounts;
			pr1->stat |= BXSTAT_LINKED;
			cclist[cccounts].boxes = 1;
			cclist[cccounts].boxlab[0] = ix;
			cclist[cccounts].xcenter = bandoff + (BANDWIDTH/2);
			cclist[cccounts].stat = 0;
			cclist[cccounts].link = -1;
			/* avheight,avheight2,angle,pos,mleft,mright
							to be set later */
			++cccounts;
		}
		pos = pr1->yoff;
		for ( j=(i+1) ; j<boxes ; j++ ){
			pr2 = &bxlst[ ipr[j] ];
			xdist = pr2->xoff - pr1->xoff;
				/* Note that xdist never be negative */
			if ( xdist == 0 )  continue;
			if ( xdist > (CWID *2) )  break;
			pos2 = pr2->yoff;
			ovl  = pr1->height;
			if ( ovl > pr2->height )  ovl = pr2->height;
			if ( ovl >= (pos  - pos2 + pr1->height)*2 )  continue;
			if ( ovl >= (pos2 - pos  + pr2->height)*2 )  continue;
			ccn = pr2->link;
			if ( ccn == -1 ){
				if ( cclist[cc0].boxes >= CCBOXES ){
					warn2 = -1;  break;
				}
				else{	cclist[cc0].boxlab[ cclist[cc0].boxes++ ] = ipr[j];
					pr2->link = cc0;
					pr2->stat |= BXSTAT_LINKED;
				}
			}
			else{	if ( ccn == cc0 )  continue;
				if ( mergeCC(cc0,ccn,cclist,bxlst) ){
					warn2 = -1;  break;
				}
			}
		}
	}
	if ( warn2 ){
		fputs("Warning: Too many PRs / CC.\n",stderr);
	}
	if ( rectlab == 0 )  delete []ipr;
	return(cccounts);
}


int CCOBJ :: checkCC(int bandoff,int cccounts,ccpac *cclist, \
			boxlist *bxlst,int peaks,peaklist *pks){
	int	ix,ix2,ccs,reason;
	int	bxn,pos,wid,angle,wsum,m_height;
	int	bias;
	for ( ix=ccs=0 ; ix<cccounts ; ix++ ){
		if ( cclist[ix].boxes == 0 )  continue;

		wsum = m_height = 0;
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			bxn = cclist[ix].boxlab[ix2];
			wid = bxlst[bxn].height;
			wsum += wid;
			if ( m_height < wid )  m_height = wid;
		}
		cclist[ix].wsum = wsum;
		cclist[ix].avheight = wsum / cclist[ix].boxes;

#ifdef _MonCC
		printf("CC %d\t[",ix);
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			printf("%d ",cclist[ix].boxlab[ix2]);
		}
		printf("]\n");
		printf("\tWidth(Max,Av) = %d,%d\t",m_height,cclist[ix].avheight);
#endif

		reason = 0;
		if ( m_height > CCHLIMIT_H ){
			reason = BXSTAT_REJECT | BXSTAT_PASS;
		}
		else{	if ( m_height < CCHLIMIT_L ){
				if ( cclist[ix].boxes < CCSLIMIT1A ){
#ifdef	_CreateAlone
					reason = BXSTAT_REJECT | BXSTAT_ALONE;
#else
					reason = BXSTAT_REJECT | BXSTAT_PASS;
#endif
				}
			}
#if 1		/* Enable this for Speed-up */
			else{	if ( cclist[ix].boxes < CCSLIMIT1 ){
#ifdef	_CreateAlone
					reason = BXSTAT_REJECT | BXSTAT_ALONE;
#else
					reason = BXSTAT_REJECT | BXSTAT_PASS;
#endif
				}
			}
#endif
		}

		if ( reason != 0 ){
			destroyCC(&cclist[ix],bxlst,reason);
#ifdef _MonCC
			printf(" ---- Deleted\n");
#endif
			continue;
		}


		lsinit(RESOLUTION_A,RESOLUTION_B,RESOLUTION_AL);
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			bxn = cclist[ix].boxlab[ix2];
			wid = bxlst[bxn].height;
			pos = bxlst[bxn].yoff + (wid /2);
			lsputdata(bxlst[bxn].xoff -bandoff -(BANDWIDTH/2),pos);
		}

		if ( cclist[ix].boxes <= 2 )	angle = 0;
		  else				angle = lsgetA();
		cclist[ix].angle   = angle;
		cclist[ix].pos = lsgetB2(angle);
#ifdef _MonCC
		printf("\tAngle = %d	Offset = %d\t",angle,cclist[ix].pos);
#endif

		if ( cclist[ix].boxes >= CCSLIMIT2 ){
			wid = cclist[ix].avheight /2;
			for ( ix2=0 ; ix2<peaks ; ix2++ ){
				if ( (cclist[ix].pos - wid) > pks[ix2].pos )  continue;
				if ( (cclist[ix].pos + wid) < pks[ix2].pos )  break;
#ifdef _MonCC
				printf("*");
#endif
				cclist[ix].stat |= CCSTAT_ATTN;
				break;
			}
		}
#ifdef _MonCC
		printf("\n");
#endif
		setCCinfo(bandoff,&cclist[ix],bxlst);

		bias = cclist[ix].avheight2 /2;
		wsum = 0;
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			bxn = cclist[ix].boxlab[ix2];
			wid = bxlst[bxn].height;
			if ( wid < bias )  wid = bias;
			wsum += wid;
		}
		cclist[ix].wsum = wsum;

	 	cclist[ (ccs++) ] = cclist[ix];
	}

	return (ccs);
}


int CCOBJ :: checkCC_2(int bandoff,int cccounts,ccpac *cclist, \
			boxlist *bxlst,int peaks,peaklist *pks){
	int	ix,ix2,ccs,reason;
	int	bxn,pos,wid,angle,wsum,m_height;
	int	bias;
	for ( ix=ccs=0 ; ix<cccounts ; ix++ ){
		if ( cclist[ix].boxes == 0 )  continue;

		wsum = m_height = 0;
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			bxn = cclist[ix].boxlab[ix2];
			wid = bxlst[bxn].height;
			wsum += wid;
			if ( m_height < wid )  m_height = wid;
		}
		cclist[ix].wsum = wsum;
		cclist[ix].avheight = wsum / cclist[ix].boxes;

#ifdef _MonCC
		printf("CC %d\t[",ix);
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			printf("%d ",cclist[ix].boxlab[ix2]);
		}
		printf("]\n");
		printf("\tWidth(Max,Av) = %d,%d\t",m_height,cclist[ix].avheight);
#endif

		reason = 0;
		if ( m_height > CCHLIMIT_H ){
			reason = BXSTAT_REJECT | BXSTAT_PASS;
		}

		if ( reason != 0 ){
			destroyCC(&cclist[ix],bxlst,reason);
#ifdef _MonCC
			printf(" ---- Deleted\n");
#endif
			continue;
		}


		lsinit(RESOLUTION_A,RESOLUTION_B,RESOLUTION_AL);
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			bxn = cclist[ix].boxlab[ix2];
			wid = bxlst[bxn].height;
			pos = bxlst[bxn].yoff + (wid /2);
			lsputdata(bxlst[bxn].xoff -bandoff -(BANDWIDTH/2),pos);
		}

		if ( cclist[ix].boxes <= 2 )	angle = 0;
		  else				angle = lsgetA();
		cclist[ix].angle   = angle;
		cclist[ix].pos = lsgetB2(angle);
#ifdef _MonCC
		printf("\tAngle = %d	Offset = %d\t",angle,cclist[ix].pos);
		printf("\n");
#endif
		setCCinfo(bandoff,&cclist[ix],bxlst);

		bias = cclist[ix].avheight2 /2;
		wsum = 0;
		for ( ix2=0 ; ix2<cclist[ix].boxes ; ix2++ ){
			bxn = cclist[ix].boxlab[ix2];
			wid = bxlst[bxn].height;
			if ( wid < bias )  wid = bias;
			wsum += wid;
		}
		cclist[ix].wsum = wsum;

	 	cclist[ (ccs++) ] = cclist[ix];
	}

	return (ccs);
}




/*------------------------------------------------------
	Band control class
------------------------------------------------------*/

class BANDCTL : public CCOBJ {
    private:
	int		bandoff;
	int		height,peaks;
	int		*hisg;
	peaklist	*pks;
	boxlist		*bxlst0;		// Box Pool
	ccpac		*cclist;		// Local CC List
	int		cccounts;

	int		rects;
	int		*rectlab;

	void		dealloc(void);
	int		nmeans(int *src,int *dst,int size,int span);
	int		getpeaklist(int *src,peaklist *pks,int size, \
					int listsize,int threshold);

    public:
	int		set_localPRs(int band,int rects0,boxlist *bxlst0);
	int		chk_histogram(int height);
	ccpac		*getcclist(void){  return cclist; }
	void		release(void){  dealloc(); }

	int		createCC(void);
	int		checkCC(void);
	int		checkCC_2(void);

			BANDCTL(void);
			~BANDCTL(void);
};


BANDCTL :: BANDCTL (void){
	hisg = 0;
	pks = 0;
	cclist = 0;
	rects = 0;
	rectlab = 0;
}


BANDCTL :: ~BANDCTL (void){
	dealloc();
}


void BANDCTL :: dealloc(void){
	if ( hisg != 0 ){  delete []hisg;  hisg = 0; }
	if ( pks != 0 ){  delete []pks;  pks = 0; }
	if ( cclist != 0 ){  delete []cclist;  cclist = 0;  cccounts = 0; }
	if ( rectlab != 0 ){  delete []rectlab;  rectlab = 0;  rects = 0; }
}


int BANDCTL :: nmeans(int *src,int *dst,int size,int span){
	int	i,j;
	int	sum,*lb;
	int	*lbuf;
	if ( NULL == (lbuf = new int [size + span] ) )  return (-1);
	for ( i=0 ; i<(span/2) ; i++ )  lbuf[i] = 0;
	for ( i= size + (span/2) ; i < size + span ; i++ )  lbuf[i] = 0;
	memcpy((char *)(lbuf+(span/2)),(char *)src,size * sizeof(int));
	for ( i=0 ; i<size ; i++ ){
		lb = &lbuf[i];
		sum = 0;
		for ( j=0 ; j<span ; j++ )  sum += *lb++;
		dst[i] = sum;
	}
	delete []lbuf;
	return (0);
}


int BANDCTL :: getpeaklist(int *src,peaklist *pks,int size,int listsize,int threshold){
	int	ix,l;
	int	dx,dx0,x0;
	int	count;
	count = 0;
	l = 0;
	x0 = dx0 = 0;
	for ( ix=0 ; ix<size ; ++ix ){
		dx = src[ix] - x0;
		if ( dx == 0 ){
			if ( dx0 > 0 )  ++l;  else  l = 0;
		}
		else{	if ( dx < 0 ){
				if ( l != 0 ){
					if ( x0 >= threshold ){
						pks->pos = ix - ((l+1)/2) -1;
						pks->value = x0;
						++pks;
						if ( (++count) >= listsize )  break;
					}
					l = 0;
				}
				else{	if ( dx0 > 0 ){
						if ( x0 >= threshold ){
							pks->pos = ix -1;
							pks->value = x0;
							++pks;
							if ( (++count) >= listsize )  break;
						}
					}
				}
			}
			else{	l = 0;
			}
			dx0 = dx;
		}
		x0 = src[ix];
	}
	return (count);
}


int BANDCTL :: set_localPRs(int band,int rects0,boxlist *bxlst0){
	int	ix,bandoff;
	int	height;
	long	pixsumratio;
	BANDCTL :: bxlst0 = bxlst0;
	bandoff = band * BANDWIDTH;
	BANDCTL :: bandoff = bandoff;
	dealloc();
	if ( 0 == (cclist = new ccpac [MAXCCS]) )  return(-1);
	if ( 0 == (rectlab = new int [rects0]) )  return(-1);
	for ( ix=0 ; ix<rects0; ix++ ){
		if ( bxlst0[ix].xoff   <  bandoff )  continue;
		if ( bxlst0[ix].xoff   >= (bandoff + BANDWIDTH) )  continue;
		height = bxlst0[ix].height;
		if ( height < LLIMIT_YSIZE )  continue;
		if ( height > ULIMIT_YSIZE )  continue;
		pixsumratio = 100 * bxlst0[ix].pixsum / (CWID * height);
		if ( (height > BPRREJ_HEIGHT) && (pixsumratio >= BPRREJ_RATIO) )  continue;
		if ( (bxlst0[ix].stat & (BXSTAT_REJECT \
			| BXSTAT_LINKED | BXSTAT_SPLITTED)) != 0 )  continue;
		rectlab[rects++] = ix;
	}
	return( rects );
}


int BANDCTL :: chk_histogram(int height){
	int	i,ix,pos;
	peaks = 0;
	BANDCTL :: height = height;
	if ( hisg == 0 ){
		if ( 0 == (hisg = new int [height]) )  return(-1);
	}
	if ( pks == 0 ){
		if ( 0 == (pks = new peaklist [height]) )  return(-1);
	}
	for ( ix=0 ; ix<height ; ix++ )  hisg[ix] = 0;
	for ( i=0 ; i<rects ; i++ ){
		ix = rectlab[i];
		pos = bxlst0[ix].yoff + (bxlst0[ix].height /2);
				/* Avoid memory crash even for illegal data */
		if ( pos < height )  hisg[ pos ] += bxlst0[ix].height;
	}
	if ( rects != 0 ){
		nmeans(hisg,hisg,height,FILTERSPAN);
		nmeans(hisg,hisg,height,FILTER2SPAN);
		peaks = getpeaklist(hisg,pks,height,height,PEAKTHRESHOLD);
	}
	else	peaks = 0;

#ifdef _MonPeak
	for ( int i=0 ; i<peaks ; i++ )
		printf("Peak pos %d\tValue  %d\n",pks[i].pos,pks[i].value);
#endif

	return(peaks);
}


int BANDCTL :: createCC(void){
	return( cccounts = CCOBJ::createCC(bandoff,rects,cclist,bxlst0,rectlab) );
}


int BANDCTL :: checkCC(void){
	cccounts = CCOBJ::checkCC(bandoff,cccounts,cclist,bxlst0,peaks,pks);
	return ( cccounts );
}


int BANDCTL :: checkCC_2(void){
	cccounts = CCOBJ::checkCC_2(bandoff,cccounts,cclist,bxlst0,peaks,pks);
	return ( cccounts );
}





/*------------------------------------------------------
	Line object control class
------------------------------------------------------*/

class LNOBJ{
    private:
	int		maxlines;

	int		ccorder[CCLABS_PAR_LN];
	void		sortCC(lnpac *lnlist,ccpac *cclist);
	int		stretch1(int lncounts,lnpac *lnlist, \
					int cccounts,ccpac *cclist0,int mode);
	int		summaryLN(int lncounts,lnpac *lnlist,ccpac *cclist);
    public:
	lnpac		*lnlist;
	int		*lnorder;

	int		alloc(int lines);

	int		lnaddcc(int lnn,int ccn,lnpac *lnlist,ccpac *cclist);
	int		mergeLN(int lnn1,int lnn2,lnpac *lnlist,ccpac *cclist);
	int		createLN(int cccounts,lnpac *lnlist,ccpac *cclist);
	int		destroyLN(lnpac *lnlist,ccpac *cclist);
	void		sortlist(int lncounts,lnpac *lnlist);
	int		stretch(int lncounts,lnpac *lnlist, \
				int cccounts,ccpac *cclist0,int times,int mode);

		/* Miscellaneous */
	void		LNinfo(int lncounts,lnpac *lnlist,ccpac *cclist);
	int		LNmonitor(int lncounts,lnpac *lnlist,ccpac *cclist, \
						boxlist *bxlst);

			LNOBJ();
	virtual		~LNOBJ();
};


LNOBJ :: LNOBJ(){
	lnlist = 0;
	lnorder = 0;
	maxlines = 0;
}


LNOBJ :: ~LNOBJ(){
	if ( lnlist != 0 )  delete []lnlist;
	if ( lnorder != 0 )  delete []lnorder;
}


int LNOBJ :: alloc(int lines){
	if ( 0 == (lnlist = new lnpac [lines]) )  return(-1);
	if ( 0 == (lnorder = new int [lines]) )  return(-1);
	maxlines = lines;
	return(0);
}


void LNOBJ :: LNinfo(int lncounts,lnpac *lnlist,ccpac *cclist){
	int	ix,ix2;
	for ( ix=0 ; ix<lncounts ; ix++ ){
		if ( lnlist[ix].ccs == 0 ) continue;
		printf("LN %d [ ",ix);
		for ( ix2=0 ; ix2<lnlist[ix].ccs ; ix2++ ){
			printf("%d ",lnlist[ix].cclab[ix2]);
		}
		printf("]\t");

		printf("(%d,%d,%d,%d)", \
			lnlist[ix].x0,lnlist[ix].y0,lnlist[ix].x1,lnlist[ix].y1);
		printf("\tWidth(Av,Av2) = %d,%d",lnlist[ix].avheight,lnlist[ix].avheight2);
		if ( lnlist[ix].ccs > 1 ){
			printf("\tAngle = %d",lnlist[ix].angle);
		}
		printf("\n");
	}
	printf("\n");
	return;
}


int LNOBJ :: LNmonitor(int lncounts,lnpac *lnlist,ccpac *cclist, \
						boxlist *bxlst){
	int	ix,ix2,ix3;
	int	ccs,ccn,bxs,bxn;
	for ( ix=0 ; ix<lncounts ; ix++ ){
		ccs = lnlist[ix].ccs;
		if ( ccs == 0 )  continue;
		printf("Line %d: \n",ix);
		for ( ix2=0 ; ix2<ccs ; ix2++ ){
			ccn = lnlist[ix].cclab[ix2];
			bxs = cclist[ccn].boxes;
			for ( ix3=0 ; ix3<bxs ; ix3++ ){
				bxn = cclist[ccn].boxlab[ix3];
				printf("\t%d(%d)",bxn,bxlst[bxn].absboxnum);
			}
		}
		printf("\n");
	}
	return(0);
}


void LNOBJ :: sortCC(lnpac *lnlist,ccpac *cclist){
	int	ix,ix2,cccounts,ccn1,ccn2;
	cccounts = lnlist->ccs;
	if ( cccounts == 0 )  return;
	for ( ix=0 ; ix<(cccounts-1) ; ix++ ){
		ccn1 = lnlist->cclab[ix];
		for ( ix2=(ix+1) ; ix2<cccounts ; ix2++ ){
			ccn2 = lnlist->cclab[ix2];
			if ( cclist[ccn1].xcenter > cclist[ccn2].xcenter ){
				lnlist->cclab[ix2] = ccn1;
				lnlist->cclab[ix ] = ccn2;
				ccn1 = ccn2;
			}
		}
	}
	--cccounts;
	lnlist->leftcc  = lnlist->cclab[0];
	lnlist->x0 = cclist[ lnlist->cclab[0] ].xcenter;
	lnlist->y0 = cclist[ lnlist->cclab[0] ].pos;
	lnlist->rightcc = lnlist->cclab[cccounts];
	lnlist->x1 = cclist[ lnlist->cclab[cccounts] ].xcenter;
	lnlist->y1 = cclist[ lnlist->cclab[cccounts] ].pos;
}


int LNOBJ :: destroyLN(lnpac *lnlist,ccpac *cclist){
	int	ix,ccn;
	for ( ix=0 ; ix<lnlist->ccs ; ix++ ){
		ccn = lnlist->cclab[ix];
		cclist[ccn].link = -1;
		cclist[ccn].stat &= (~ CCSTAT_LINKED);
	}
	lnlist->stat = 0;
	lnlist->ccs = 0;
	return(0);
}


/*------------------------------  Not included in class LNOBJ. */
static ccpac *cclist_forcmp;

static int cmp_ccpos(const void *n1,const void *n2){
	return( cclist_forcmp[*(int *)n1].xcenter - cclist_forcmp[*(int *)n2].xcenter );
}
/*-------------------------------------------------------------*/


int LNOBJ :: createLN(int cccounts,lnpac *lnlist,ccpac *cclist){
	int	ix,ix2,ic,ic2,lccs,ccn,cc0,lncounts,warn;
	int	angle,angle2,wid,wid2,ovl,pos,pos2,ag0;
	int	*lcc;
	if ( 0 == (lcc = new int [cccounts]) )  return(-1);
	for ( ix=lccs=0 ; ix<cccounts ; ix++ ){
		if ( (cclist[ix].stat & CCSTAT_ATTN) == 0 )  continue;
		lcc[ lccs++ ] = ix;
	}
	cclist_forcmp = cclist;
	qsort((void *)lcc,lccs,sizeof(int),cmp_ccpos);

	warn = -1;
	lncounts = 0;
	for ( ix=0 ; ix<lccs ; ix++ ){
		ic = lcc[ix];
		if ( (cclist[ic].stat & CCSTAT_LINKED) == 0 ){
			if ( lncounts >= MAXLINES ){
				if ( warn ){
					warn = 0;
					fprintf(stderr,"Warning: Too many lines\n");
				}
				break;
			}
			cclist[ic].stat |= CCSTAT_LINKED;
			cclist[ic].link = lncounts;
			lnlist[lncounts].stat = 0;
			lnlist[lncounts].ccs = 1;
			lnlist[lncounts].cclab[0] = ic;
			/* x0,y0,x1,y1,leftcc,rightcc,avheight,avheight2,angle
								to be set later */
			++lncounts;
		}
		cc0 = cclist[ic].link;
		angle = cclist[ic].angle;
		ag0 = (angle * (BANDWIDTH/2)) /RESOLUTION_A;
		for ( ix2=(ix+1) ; ix2<lccs ; ix2++ ){
			ic2 = lcc[ix2];
			ovl = cclist[ic2].xcenter - cclist[ic].xcenter;
				/* Note that ovl never be negative */
			if ( ovl > BANDWIDTH )  break;
			if ( ovl != BANDWIDTH )  continue;
			angle2 = cclist[ic2].angle;
			if ( ovl < 0 ){
				if ( (cclist[ic].mright + cclist[ic2].mleft) >= 2 ){
					continue;
				}
				pos = cclist[ic].pos + ag0;
				pos2 = cclist[ic2].pos \
					- (angle2 * (BANDWIDTH/2)) /RESOLUTION_A ;
			}
			else{	if ( (cclist[ic].mleft + cclist[ic2].mright) >= 2 ){
					continue;
				}
				pos = cclist[ic].pos - ag0;
				pos2 = cclist[ic2].pos \
					+ (angle2 * (BANDWIDTH/2)) /RESOLUTION_A;
			}
			wid  = cclist[ic ].avheight2;
			wid2 = cclist[ic2].avheight2;
						//   (average of wid & wid2) / 2
			if ( ((wid + wid2) / 4) < abs(pos - pos2) ){
				continue;
			}
			if ( (wid  * 3) < wid2 )  continue;
			if ( (wid2 * 3) < wid  )  continue;
			if ( (cclist[ic2].stat & CCSTAT_LINKED) == 0 ){
				lnaddcc(cc0,ic2,lnlist,cclist);
			}
			else{	ccn = cclist[ic2].link;
				if ( ccn == cc0 )  continue;
				mergeLN(cc0,ccn,lnlist,cclist);
			}
		}
	}
	summaryLN(lncounts,lnlist,cclist);
	delete []lcc;
	return(lncounts);
}


int LNOBJ :: lnaddcc(int lnn,int ccn,lnpac *lnlist,ccpac *cclist){
	int	ccx;
	if ( lnlist[lnn].ccs >= CCLABS_PAR_LN ){
		fprintf(stderr,"Warning: Too many CCs / line.\n");
		return(-1);
	}
	lnlist[lnn].cclab[ lnlist[lnn].ccs++ ] = ccn;
	cclist[ccn].link = lnn;
	cclist[ccn].stat |= CCSTAT_LINKED;
	ccx = cclist[ccn].xcenter;
	if ( lnlist[lnn].x0 > ccx ){
		lnlist[lnn].x0 = ccx;
		lnlist[lnn].y0 = cclist[ccn].pos;
		lnlist[lnn].leftcc  = ccn;
	}
	if ( lnlist[lnn].x1 < ccx ){
		lnlist[lnn].x1 = ccx;
		lnlist[lnn].y1 = cclist[ccn].pos;
		lnlist[lnn].rightcc  = ccn;
	}
	return(0);
}


int LNOBJ :: mergeLN(int lnn1,int lnn2,lnpac *lnlist,ccpac *cclist){
	int	ix,ccn,ccx;
	int	ccs1,ccs2;
	ccs1 = lnlist[lnn1].ccs;
	ccs2 = lnlist[lnn2].ccs;
	if ( (ccs1 + ccs2) > CCLABS_PAR_LN ){
		fprintf(stderr,"Warning: Too many CCs / line.\n");
		return(-1);
	}
	for ( ix=0 ; ix<ccs2 ; ix++ ){
		ccn = lnlist[lnn2].cclab[ix];
		cclist[ccn].link = lnn1;
		lnlist[lnn1].cclab[ ccs1++ ] = ccn;
		ccx = cclist[ccn].xcenter;
		if ( lnlist[lnn1].x0 > ccx ){
			lnlist[lnn1].x0 = ccx;
			lnlist[lnn1].y0 = cclist[ccn].pos;
			lnlist[lnn1].leftcc  = ccn;
		}
		if ( lnlist[lnn1].x1 < ccx ){
			lnlist[lnn1].x1 = ccx;
			lnlist[lnn1].y1 = cclist[ccn].pos;
			lnlist[lnn1].rightcc  = ccn;
		}
	}
	lnlist[lnn1].stat |= (lnlist[lnn2].stat & LNSTAT_ATTN);
	lnlist[lnn1].ccs = ccs1;
	lnlist[lnn2].ccs = 0;
	return(0);
}


int LNOBJ :: summaryLN(int lncounts,lnpac *lnlist,ccpac *cclist){
	int	ix,ix2;
	int	ccn,angle,avw,avw2,bsum,csum,xdif;
	for ( ix=0 ; ix<lncounts ; ix++ ){
		if ( lnlist[ix].ccs == 0 ) continue;
		sortCC(&lnlist[ix],cclist);
		avw = avw2 = 0;
		bsum = csum = 0;
		for ( ix2=0 ; ix2<lnlist[ix].ccs ; ix2++ ){
			ccn = lnlist[ix].cclab[ix2];
			bsum += cclist[ccn].boxes;
			csum += cclist[ccn].columns;
			avw  += cclist[ccn].wsum;
			avw2 += cclist[ccn].wsum2;
		}
		lnlist[ix].avheight  = avw  / bsum;
		lnlist[ix].avheight2 = avw2 / csum;
		lnlist[ix].wsum  = avw;
		lnlist[ix].bsum  = bsum;
		lnlist[ix].wsum2 = avw2;
		lnlist[ix].csum  = csum;
		xdif = lnlist[ix].x1 - lnlist[ix].x0;
		if ( xdif != 0 ){
//		if ( lnlist[ix].ccs > 1 ){
			angle = ((lnlist[ix].y1 - lnlist[ix].y0)*RESOLUTION_A) / xdif;
		}
		else{	angle = cclist[0].angle;
		}
		lnlist[ix].angle = angle;
	}
	return (0);
}


/*------------------------------  Not included in class LNOBJ. */
static lnpac	*lnlist_forcmp;

static int cmp_linelength(const void *c1,const void *c2){
	return ( lnlist_forcmp[ *(int *)c2 ].ccs - lnlist_forcmp[ *(int *)c1 ].ccs );
}
/*-------------------------------------------------------------*/


void LNOBJ :: sortlist(int lncounts,lnpac *lnlist){
	int	ix;
	lnlist_forcmp = lnlist;
	for ( ix=0 ; ix<lncounts ; ix++ )  lnorder[ix] = ix;
	qsort((void *)lnorder,lncounts,sizeof(int),cmp_linelength);
#ifdef _MonPriority
	printf("Line priority order :\n");
	for ( ix=0; ix<lncounts ; ix++ ){
		if ( lnlist[ lnorder[ix] ].ccs == 0 )  break;
		printf("%d ",lnorder[ix]);
	}
	printf("\n");
#endif
}


int LNOBJ :: stretch1(int lncounts,lnpac *lnlist, \
				int cccounts,ccpac *cclist0,int mode){
	int	ix,ix2,ix3,ccn,lnlab,lnlab2,retry;
	int	ovl,angle,angle2,pos,pos2,wid,wid2,xc;

	sortlist(lncounts,lnlist);

	retry = 0;
	for ( ix=0 ; ix<lncounts ; ix++ ){
		lnlab = lnorder[ix];
		if ( (lnlist[lnlab].stat & LNSTAT_ATTN) == 0 )  continue;
		lnlist[lnlab].stat &= (~ LNSTAT_ATTN);
		if ( mode == 0 ){
			if ( lnlist[lnlab].ccs < 2 )  break;  
		}
#ifdef _MonLNLink
		printf("LN %d:\t",lnlab);
#endif
		angle = lnlist[lnlab].angle;
		for ( ix2=0 ; ix2<2 ; ix2++ ){
			if ( ix2 == 0 ){
				ccn = lnlist[lnlab].leftcc;
				xc  = cclist0[ccn].xcenter - BANDWIDTH;
				pos = cclist0[ccn].pos \
					- (angle * (BANDWIDTH/2)) /RESOLUTION_A;
				ovl = 1;
			}
			else{	ccn = lnlist[lnlab].rightcc;
				xc  = cclist0[ccn].xcenter + BANDWIDTH;
				pos = cclist0[ccn].pos \
					+ (angle * (BANDWIDTH/2)) /RESOLUTION_A ;
				ovl = -1;
			}
			for ( ix3=0 ; ix3<cccounts ; ix3++ ){
				if ( cclist0[ix3].xcenter != xc )  continue;
				if ( cclist0[ix3].boxes == 0 )  continue;
				angle2 = cclist0[ix3].angle;
				if ( ovl < 0 ){
					if ( (cclist0[ccn].mright + cclist0[ix3].mleft) >= 2 ){
						continue;
					}
					pos2 = cclist0[ix3].pos \
						- (angle2 * (BANDWIDTH/2)) /RESOLUTION_A ;
				}
				else{	if ( (cclist0[ccn].mleft + cclist0[ix3].mright) >= 2 ){
						continue;
					}
					pos2 = cclist0[ix3].pos \
						+ (angle2 * (BANDWIDTH/2)) /RESOLUTION_A;
				}
				wid  = lnlist[lnlab].avheight2;
				wid2 = cclist0[ix3].avheight2;
				if ( (wid  * 4) < wid2 )  continue;
				if ( (wid2 * 4) < wid  )  continue;
						//   (average of wid & wid2) / 2
				if ( ! mode ){
					if ( ((wid + wid2) / 4) < abs(pos - pos2) ){
						continue;
					}
				}
				else{
					if ( ((wid + wid2) / 2) < abs(pos - pos2) ){
						continue;
					}
				}
				if ( (cclist0[ix3].stat & CCSTAT_LINKED) != 0 ){
					lnlab2 = cclist0[ix3].link;
					if ( lnlab2 != lnlab ){
						if ( ovl < 0 ){
							if( ix3 != lnlist[lnlab2].leftcc )  continue;
						}
						else{	if( ix3 != lnlist[lnlab2].rightcc )  continue;
						}
						retry = -1;
						lnlist[lnlab].stat |= LNSTAT_ATTN;
#ifdef _MonLNLink
						printf("[LN:%d]",lnlab2);
#endif
						mergeLN(lnlab,lnlab2,lnlist,cclist0);
					}
					continue;
				}
				else{	lnaddcc(lnlab,ix3,lnlist,cclist0);
					retry = -1;
					lnlist[lnlab].stat |= LNSTAT_ATTN;
#ifdef _MonLNLink
					printf("[CC:%d]",ix3);
#endif
				}
			}
		}
#ifdef _MonLNLink
		printf("\n\n");
#endif
	}

	summaryLN(lncounts,lnlist,cclist0);
#ifdef _MonLNInfo
	LNinfo(lncounts,lnlist,cclist0);
#endif

	return ( retry );
}


int LNOBJ :: stretch(int lncounts,lnpac *lnlist, \
			int cccounts,ccpac *cclist0,int times,int mode){
	int	ix,pass;
	for ( ix=0 ; ix<lncounts ; ix++ )  lnlist[ix].stat |= LNSTAT_ATTN;
	for ( pass=0 ; pass<times ; pass++ ){
		if ( ! stretch1(lncounts,lnlist, \
					cccounts,cclist0,mode) )  break;
	}
	return(0);
}




/*------------------------------------------------------
	CC's pool class
------------------------------------------------------*/

class CCPOOL {
    private:
	ccpac		*pool;
    public:
	int		cccounts;

	void		dealloc(void);
	int		addccs(int ccs,ccpac *cclist);
	ccpac		*getpool(void) {  return (pool); }
	int		clear(void){  dealloc();  return(0); }

		/* Miscellaneous */
	void		print_CCinfo(void);

			CCPOOL(void);
	virtual		~CCPOOL(void);
};


CCPOOL :: CCPOOL (void){
	pool = 0;
	cccounts = 0;
}


CCPOOL :: ~CCPOOL (void){
	dealloc();
}


void CCPOOL :: dealloc(void){
	if ( pool != 0 ){  delete []pool;  pool = 0;  cccounts = 0; }
}


int CCPOOL :: addccs (int ccs,ccpac *cclist){
	int	ix;
	if ( pool == 0 ){
		if ( 0 == (pool = new ccpac[MAXGLCCS]) )  return(-1);
	}
	if ( (cccounts + ccs) > MAXGLCCS )  return(-1);
	for ( ix=0 ; ix<ccs ; ix++ ){
		if ( cclist[ix].boxes == 0 )  continue;
		pool[ (cccounts++) ] = cclist[ix];
	}
	return(0);
}



/*----------------------------------  Macro functions */

void CCPOOL :: print_CCinfo(void){
	int	ix,pos,pos2,wid,wid2,angle;
	for ( ix=0 ; ix<cccounts ; ix++ ){
		pos2 = pool[ix].xcenter;
		pos  = pool[ix].pos;
		angle = pool[ix].angle;
		wid  = pool[ix].avheight;
		wid2 = pool[ix].avheight2;
		printf("CC %d (%d)\tPos = (%d,%d)\tAngle = %d\tWidth(Av,Av2) = %d,%d\n", \
					ix,pool[ix].boxes,pos2,pos,angle,wid,wid2);
	}
}




/*------------------------------------------------------
	String candidate export class
------------------------------------------------------*/

class STREXP : public D4SAVE{
    private:
	d4file		d4header;
	d4pac		d4p;
	int		*lntop,*lnbottom;
	int		columns;
	void		release(void);
	int		open(char *fname,int x0,int x1,int y0,int y1, \
				int width0,int width1,int height0,int height1);
	int		close(void);
    public:
	int		write(char *fname,int x0,int x1,int y0,int y1, \
				int width0,int width1,int height0,int height1, \
				int lncounts,lnpac *lnlist,ccpac *cclist0, \
								boxlist *bxlst0);
			STREXP(void);
			~STREXP(void);
};


STREXP :: STREXP (void){
	lntop = lnbottom = 0;
}


STREXP :: ~STREXP (void){
	release();
}


void STREXP :: release(void){
	if ( lntop != 0 )  delete []lntop;
	lntop = 0;
	if ( lnbottom != 0 )  delete []lnbottom;
	lnbottom = 0;
}


int STREXP :: write(char *fname,int x0,int x1,int y0,int y1, \
			int width0,int width1,int height0,int height1, \
			int lncounts,lnpac *lnlist,ccpac *cclist0, \
							boxlist *bxlst0){
	int	ix,ix2,ix3;
	int	ccs,ccn,bxs,bxn,mxh,lnn;
	int	bx0,by0,by1,col,lntotal,exsw;
	int	vaboxes,boxes,bleft,bright;
	int	lines;

	typedef struct {
		int	id;
		int	val;
	} sortlist;

	sortlist	*order;
	sortlist	otmp;

	columns = (x1 + (CWID-1)) /CWID;
	release();
	lntop = new int[columns];
	lnbottom = new int[columns];
	order = new sortlist[lncounts];
	if ( (lntop == 0) || (lnbottom == 0) || (order == 0) ){
		fprintf(stderr,"STREXP : Memory allocation failed.\n");
		return (-1);
	}
	if ( 0 != open(fname,x0,x1,y0,y1,width0,width1,height0,height1) ){
		fprintf(stderr,"File creation failed : %s\n",fname);
		return (-1);
	}
	for ( ix=lines=0 ; ix<lncounts ; ix++ ){
		ccs = lnlist[ix].ccs;
		if ( ccs == 0 )  continue;
		order[lines].id = ix;
		order[lines].val = (cclist0[lnlist[ix].cclab[0]].pos \
				+ cclist0[lnlist[ix].cclab[ccs-1]].pos) /2;
		++lines;
	}
	for ( ix=0 ; ix<(lines-1) ; ix++ ){
		for ( ix2=ix+1 ; ix2<lines ; ix2++ ){
			if ( order[ix].val > order[ix2].val ){
				otmp = order[ix];
				order[ix] = order[ix2];
				order[ix2] = otmp;
			}
		}
	}
	for ( ix=0 ; ix<lines ; ix++ ){
		lnn = order[ix].id;
		ccs = lnlist[lnn].ccs;
		if ( ccs == 0 )  continue;
		for ( ix2=0 ; ix2<columns ; ix2++ ){	// Clear string buffer
			lntop[ix2] = 32767;
			lnbottom[ix2] = -1;
		}
		for ( ix2=0 ; ix2<ccs ; ix2++ ){
			ccn = lnlist[lnn].cclab[ix2];
			bxs = cclist0[ccn].boxes;
			for ( ix3=0 ; ix3<bxs ; ix3++ ){
				bxn = cclist0[ccn].boxlab[ix3];
				bx0 = bxlst0[bxn].xoff;
				col = bx0 / CWID;
				by0 = bxlst0[bxn].yoff;
				by1 = by0 + bxlst0[bxn].height -1;
				if ( lntop[col] > by0 )  lntop[col] = by0;
				if ( lnbottom[col] < by1 )  lnbottom[col] = by1;
			}
		}

		mxh = 0;
		bleft = -1;	bright = 0;
		exsw = -1;
		for ( ix2=vaboxes=0 ; ix2<columns ; ix2++ ){	// Get maximum height
			bx0 = lnbottom[ix2] - lntop[ix2] +1;
			if ( bx0 <= 0 )  continue;
			if ( mxh < bx0 )  mxh = bx0;
			if ( bleft == -1 )  bleft = ix2;
			bright == ix2;
			++vaboxes;
		}
		boxes = bright - bleft +1;

//		if ( (mxh >= LNHLIMIT) || (mxh <= LNLLIMIT) ){
		if ( (lnlist[lnn].avheight2 >= LNHLIMIT) || (mxh <= LNLLIMIT) ){
			lnlist[lnn].stat |= LNSTAT_KILL;
			exsw = 0;
		}

		if ( vaboxes <= (NBX -1) ){
			if ( (vaboxes <= (NBX -2)) \
				|| (boxes > ((NBX -1) * CWID)) ){
				lnlist[lnn].stat |= LNSTAT_KILL;
				exsw = 0;
			}
		}

		for ( ix2=lntotal=0 ; (ix2<columns) && exsw ; ix2++ ){
			bx0 = lnbottom[ix2] - lntop[ix2] +1;
			if ( bx0 <= 0 )  continue;
			d4p.d4_data[0] = ix2 * CWID;
			d4p.d4_data[1] = lntop[ix2];
			d4p.d4_data[2] = CWID;
			d4p.d4_data[3] = bx0;
			if ( 0 != putdata(&d4p) ){
				fprintf(stderr,"STREXP : File write error : %s\n",fname);
				close();  unlink(fname);
				return (-1);
			}
			++lntotal;
		}
		if ( lntotal != 0 ){
			d4p.d4_data[0] = d4p.d4_data[1] = 0;
			d4p.d4_data[2] = d4p.d4_data[3] = 0;
			if ( 0 != putdata(&d4p) ){
				fprintf(stderr,"STREXP : File write error : %s\n",fname);
				close();  unlink(fname);
				return (-1);
			}
		}
	}
	if ( 0 != close() ){
		fprintf(stderr,"STREXP : File write error : %s\n",fname);
		return (-1);
	}
	return(0);
}


int STREXP :: open(char *fname,int x0,int x1,int y0,int y1, \
			int width0,int width1,int height0,int height1){
	d4header.d4_llimit[0] = x0;		d4header.d4_hlimit[0] = x1;
	d4header.d4_llimit[1] = y0;		d4header.d4_hlimit[1] = y1;
	d4header.d4_llimit[2] = width0;		d4header.d4_hlimit[2] = width1;
	d4header.d4_llimit[3] = height0;	d4header.d4_hlimit[3] = height1;
	return ( filecreat(fname,&d4header) );
}


int STREXP :: close(void){
	return ( fileclose() );
}




/*------------------------------------------------------
	Box list file export class
------------------------------------------------------*/

class BXEXP {
    private:
	D4SAVE		D4SV;
	D8SAVE		D8SV;
	int		ndim;
	int		open(char *fname,int x0,int x1,int y0,int y1, \
				int width0,int width1,int height0,int height1);
	int		close(void);
    public:
	int		writebxdata(char *fname,int x0,int x1,int y0,int y1, \
				int width0,int width1,int height0,int height1, \
				int boxes,boxlist *bxlst0,int mask,int cond,int ndim);
};


int BXEXP :: writebxdata(char *fname,int x0,int x1,int y0,int y1, \
			int width0,int width1,int height0,int height1, \
			int boxes,boxlist *bxlst0,int mask,int cond,int ndim){
	int	ix,stat;
	long	pixsum;
	d4pac	d4p;
	d8pac	d8p;
	d8p.d8_data[6] = d8p.d8_data[7] = 0;
	BXEXP::ndim = ndim;
	if ( 0 != open(fname,x0,x1,y0,y1,width0,width1,height0,height1) ){
		fprintf(stderr,"File creation failed : %s\n",fname);
		return (-1);
	}
	for ( ix=0 ; ix<boxes ; ix++ ){
		stat = bxlst0[ix].stat;
		if ( (stat & BXSTAT_LINKED) != 0 )  continue;
//		if ( (stat & BXSTAT_CLUSTER) != 0 )  continue;
		stat &= mask;
		if ( (stat ^ cond) != 0 )  continue;

		if ( ndim == 4 ){
			d4p.d4_data[0] = bxlst0[ix].xoff;
			d4p.d4_data[1] = bxlst0[ix].yoff;
			d4p.d4_data[2] = bxlst0[ix].width;
			d4p.d4_data[3] = bxlst0[ix].height;
			if ( 0 != D4SV.putdata(&d4p) ){
				fprintf(stderr,"File write error : %s\n",fname);
				close();
				unlink(fname);
				return (-1);
			}
		}
		else{
			d8p.d8_data[0] = bxlst0[ix].xoff;
			d8p.d8_data[1] = bxlst0[ix].yoff;
			d8p.d8_data[2] = bxlst0[ix].width;
			d8p.d8_data[3] = bxlst0[ix].height;
			pixsum = bxlst0[ix].pixsum;
			d8p.d8_data[4] = (short)(pixsum >> 16);
			d8p.d8_data[5] = (short)(pixsum & 0xffffL);
			if ( 0 != D8SV.putdata(&d8p) ){
				fprintf(stderr,"File write error : %s\n",fname);
				close();
				unlink(fname);
				return (-1);
			}
		}
	}
	if ( 0 != close() ){
		fprintf(stderr,"File write error : %s\n",fname);
		return (-1);
	}
	return(0);
}


int BXEXP :: open(char *fname,int x0,int x1,int y0,int y1, \
			int width0,int width1,int height0,int height1){
	d4file	d4header;
	d8file	d8header;
	if ( ndim == 4 ){
		d4header.d4_llimit[0] = x0;		d4header.d4_hlimit[0] = x1;
		d4header.d4_llimit[1] = y0;		d4header.d4_hlimit[1] = y1;
		d4header.d4_llimit[2] = width0;		d4header.d4_hlimit[2] = width1;
		d4header.d4_llimit[3] = height0;	d4header.d4_hlimit[3] = height1;
		return ( D4SV.filecreat(fname,&d4header) );
	}
	else{
		d8header.d8_llimit[0] = x0;		d8header.d8_hlimit[0] = x1;
		d8header.d8_llimit[1] = y0;		d8header.d8_hlimit[1] = y1;
		d8header.d8_llimit[2] = width0;		d8header.d8_hlimit[2] = width1;
		d8header.d8_llimit[3] = height0;	d8header.d8_hlimit[3] = height1;
		d8header.d8_llimit[4] = 0;		d8header.d8_hlimit[4] = 0x05;
		d8header.d8_llimit[5] = 0;		d8header.d8_hlimit[5] = -1;
		d8header.d8_llimit[6] = 0;		d8header.d8_hlimit[6] = 0;
		d8header.d8_llimit[7] = 0;		d8header.d8_hlimit[7] = 0;
	}
	return ( D8SV.creat(fname,&d8header) );
}


int BXEXP :: close(void){
	if ( ndim == 4 )  return ( D4SV.fileclose() );
	return( D8SV.close() );
}




/*------------------------------------------------------
	postscript file export class
------------------------------------------------------*/

class PSEXP {
    private:
	FILE		*fp;
    public:
	int		open(FILE *fp);
	int		close(void);

	int		line(int x0,int y0,int x1,int y1,int width);
	int		line0(int x0,int y0,int width);
	int		lineto(int x,int y);
	int		stroke(void);
	int		marker(int x,int y);
	int		writelndata(char *fname,int lncounts,\
					lnpac *lnlist,ccpac *cclist0);
			PSEXP(void);
};


PSEXP :: PSEXP(void){
	fp = stdout;
}



int PSEXP :: open(FILE *fp){
	PSEXP :: fp = fp;
	stdps_setsize("A4");
	return( stdps_header(fp) );
}


int PSEXP :: close(void){
	return( stdps_end(fp) );
}


int PSEXP :: line0(int x0,int y0,int width){
	fprintf(fp,"%d setlinewidth\n",width);
	fprintf(fp,"newpath %d %d moveto\n",x0,y0);
	return(0);
}


int PSEXP :: lineto(int x,int y){
	fprintf(fp,"	%d %d lineto\n",x,y);
	return(0);
}


int PSEXP :: marker(int x,int y){
	fprintf(fp,"1 setlinewidth\n");
	fprintf(fp,"newpath %d %d moveto ",x-3,y-3);
	fprintf(fp,"%d %d lineto ",x+3,y-3);
	fprintf(fp,"%d %d lineto ",x+3,y+3);
	fprintf(fp,"%d %d lineto closepath stroke\n",x-3,y+3);
	return(0);
}


int PSEXP :: stroke(void){
	fprintf(fp,"	stroke\n");
	return(0);
}


int PSEXP :: line(int x0,int y0,int x1,int y1,int width){
	fprintf(fp,"%d setlinewidth\n",width);
	fprintf(fp,"newpath %d %d moveto %d %d lineto  stroke\n",x0,y0,x1,y1);
	return(0);
}


int PSEXP :: writelndata(char *fname,int lncounts,lnpac *lnlist,ccpac *cclist0){
	int	ix,ix2,ccn,pos,pos2;
	FILE	*fp;
	if ( strcmp(fname,"stdout") == 0 ){
		fp = stdout;  open(fp);
	}
	else{	if ( 0 == (fp = fopen(fname,"w")) )  return(-1);
		if ( open(fp) ){  fclose(fp);  unlink(fname);  return(-1); }
	}
	for ( ix=0; ix<lncounts ; ix++ ){
		if ( lnlist[ix].ccs == 0 )  continue;

		ccn = lnlist[ix].cclab[0];
		if ( lnlist[ix].ccs == 1 ){
			marker(cclist0[ccn].xcenter,cclist0[ccn].pos);
			continue;
		}
		line0(cclist0[ccn].xcenter,cclist0[ccn].pos,3);
		for ( ix2=1 ; ix2<lnlist[ix].ccs ; ix2++ ){
			ccn = lnlist[ix].cclab[ix2];
			pos = cclist0[ccn].pos;
			pos2 = cclist0[ccn].xcenter;
			lineto(pos2,pos);
		}
		stroke();
	}
	if ( close() ){
		if ( fp != stdout )  unlink(fname);
	}
	return(0);
}




/*------------------------------------------------------
	Message class
------------------------------------------------------*/

class MSG {
    public:
	void		usage(void);
	void		switcherr(void);
	void		readerr(char *name);
	void		writeerr(char *name);
	void		nonsupported(char *name);
	void		memoryover(void);
};
    	
    	
void MSG :: usage(void){		// Display Usage
	puts("lnextr2  v2.0, Copyright (C) 1994-1997 Hideaki Goto ");
	puts("Usage : lnextr2 [-options] in_file ");
	puts("	      	-rjpass n  : rejection passes (def. n=4) ");
	puts("	      	-split     : split rejected PRs ");
	puts("	      	-splits n  : splitting flagments (def. n=4) ");
	puts("	      	-o   fname : output Line-file name ");
//	puts("	      	-pr  fname : output PR-file name ");
	puts("	      	-rpr fname : output residue-PR-file name ");
//	puts("	      	-apr fname : output alone-PR-file name ");
	puts("	      	-ps  fname : output postscript-info-file name ");
	puts("	      	-info      : print processing information ");
}


void MSG :: switcherr(void){
	puts("Switch error.");
}


void MSG :: readerr(char *name){
	printf("Can't read %s.\n",name);
}


void MSG :: writeerr(char *name){
	printf("Can't write %s.\n",name);
}


void MSG :: nonsupported(char *name){
	printf("Non-supported file %s.\n",name);
}


void MSG :: memoryover(void){
	puts("Not enough memory.");
}




/*------------------------------------------------------
	Argument / Option class
------------------------------------------------------*/

class ARGS {
    private:
	char		*fpath;
	int		fpcount;
	int		sw_split;
	int		sw_splits;
	int		sw_rjpass;
	int		sw_rjsusp;
	int		sw_info;
	char		*fn_rpr,*fn_apr,*fn_ps;
	char		*fn_out;
    public:
	int		scanargv(int ac,char **av);
	int		SWsplit(void) {  return sw_split; }
	int		SWsplits(void) {  return sw_splits; }
	int		SWrjpass(void) {  return sw_rjpass; }
	int		SWrjsusp(void) {  return sw_rjsusp; }
	int		SWinfo(void) {  return sw_info; }
	char		*fname_rpr(void) {  return fn_rpr; }
	char		*fname_apr(void) {  return fn_apr; }
	char		*fname_ps(void) {  return fn_ps; }
	char		*infile(void) { return fpath; }
	char		*outfile(void) {  return fn_out; }
			ARGS(void);
};


ARGS :: ARGS(void){
	fpath = "";
	sw_split = 0;
	sw_splits = 4;
	sw_rjpass = 4;
	sw_rjsusp = 0;
	sw_info = 0;
	fn_rpr = fn_apr = fn_ps = 0;
	fn_out = "/dev/null";
}


int ARGS :: scanargv(int ac,char **av){
	int	i,err,skip;
	char	*sw;
	fpcount = err = 0;
	skip = 0;
	for( i = 1 ; i < ac ; i++ ){
		if( skip ){  skip = 0;  continue;  }
		if( *( sw = av[i] ) == '-' ){
			++sw;
			if( 0 == strcmp(sw,"rjpass") ){
				if(i<ac){
					sw_rjpass = atoi(av[i+1]);
					if ( sw_rjpass == 0 )  sw_rjsusp = -1;
					if ( sw_rjpass < 1 )  sw_rjpass = 1;
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"split") ){
				sw_split = -1;
				continue;
			}
			if( 0 == strcmp(sw,"splits") ){
				if(i<ac){
					sw_splits = atoi(av[i+1]);
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"info") ){
				sw_info = -1;
				continue;
			}
			if( 0 == strcmp(sw,"o") ){
				if(i<ac){
					fn_out = av[i+1];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"rpr") ){
				if(i<ac){
					fn_rpr = av[i+1];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"apr") ){
				if(i<ac){
					fn_apr = av[i+1];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"ps") ){
				if(i<ac){
					fn_ps = av[i+1];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			err = -1;  break;
		}
		else{	if ( fpcount == 0 ){ fpath = av[i];  fpcount++; }
		}
	}
	return(err);
}




/*------------------------------------------------------
	Main routine
------------------------------------------------------*/

int main(int ac,char **av){
	ARGS		ARGS;
	MSG		MSG;
	D8LOAD		D8LD;
	CCPOOL		CCPOOL;
	BXOBJ		BXOBJ;		// Global Box Object
	CCOBJ		CCOBJ;		// Global CC Object
	BANDCTL		BANDCTL;	// Band Object (with Local Box/CC Object)
	LNOBJ		LNOBJ;
	PSEXP		PSEXP;
	BXEXP		BXEXP;
	STREXP		STREXP;
	int		width,height;
	int		retcode;
	int		ix,boxes;
	D8Rect		d8pr;
	boxlist		*bxlst0,box;
	ccpac		*cclist0;
	int		cccounts;
	lnpac		*lnlist;
	int		lncounts;
	int		boxcounter;
	int		bandoff,pass;
	int		peaks;

	if ( ac < 2 ){  MSG.usage();  return(0);  }

	if ( 0 != ARGS.scanargv(ac,av) ){
		MSG.switcherr();
		exit(1);
	}

	if ( LNOBJ.alloc(MAXLINES) ){
		fputs("Memory not enough.\n",stderr);
		return(1);
	}
	lnlist = LNOBJ.lnlist;




//------------------------------------- File read phase


	retcode = D8LD.open(ARGS.infile());
	switch ( retcode ){
	  case 0:	break;
	  case -3:	MSG.nonsupported(ARGS.infile());  return(2);
	  case -4:	MSG.memoryover();  return(3);
	  default:	MSG.readerr(ARGS.infile());  return(2);
	}

	D8LD.getupper((d8pac *)(&d8pr));
	width  = (int)d8pr.x0 + BANDWIDTH;
	width  = (width /BANDWIDTH) *BANDWIDTH;
	height = (int)d8pr.y0 +1;

	boxcounter = boxes = 0;
	D8LD.rewind();
	while ( 0 == D8LD.getdata((d8pac *)(&d8pr)) ){
		++boxcounter;
		if ( d8pr.h > ULIMIT_YSIZE0 )  continue;
		if ( d8pr.h < LLIMIT_YSIZE0 )  continue;
		box.absboxnum = boxcounter -1;
		box.xoff  = d8pr.x0;	box.yoff   = d8pr.y0;
		box.width = d8pr.w;	box.height = d8pr.h;
		box.pixsum = d8pr.pixsum;
		if ( (boxes = BXOBJ.addnewbox(&box,0)) < 0 ){
			fputs("Memory not enough.\n",stderr);
			return(1);
		}
	}

	D8LD.close();
	if ( ARGS.SWinfo() ){
		printf ("Total           rectangles = %d\n",boxcounter);
//		printf ("Rejected (huge) rectangles = %d\n",boxcounter - boxes);
		printf ("\n");
	}


	CCPOOL.clear();




//------------------------------------- Line extraction phase


    for ( pass=0 ; pass<ARGS.SWrjpass() ; pass++ ){

	if ( ARGS.SWsplit() ){
		if ( 0 != BXOBJ.splitall(ARGS.SWsplits()) ){
			printf("Warning: split error.\n");
		}
	}

	boxcounter = BXOBJ.getboxes();
	bxlst0 = BXOBJ.getbuf();
	for ( bandoff=0 ; bandoff<width ; bandoff+=BANDWIDTH ){
		if ( ARGS.SWinfo() ){
			printf("------------------------ Band %d ------------------------\n",bandoff/BANDWIDTH);
		}

		boxes = BANDCTL.set_localPRs(bandoff/BANDWIDTH,boxcounter,bxlst0);
		if ( boxes < 0 ){
			fputs("Memory not enough.\n",stderr);
			return(1);
		}
		if ( ARGS.SWinfo() )  printf ("Available boxes = %d\n",boxes);

		if ( (peaks = BANDCTL.chk_histogram(height)) < 0 ){
			fputs("Memory not enough.\n",stderr);
			return(1);
		}
		if ( ARGS.SWinfo() )  printf ("Available peaks = %d\n",peaks);

		cccounts = BANDCTL.createCC();
		cccounts = BANDCTL.checkCC();

		if ( ARGS.SWinfo() )  printf("Available CCs = %d\n",cccounts);
		if ( 0 != CCPOOL.addccs(cccounts,BANDCTL.getcclist()) ){
			fprintf(stderr,"Warning: CCs pool overflow\n");
		}

		BANDCTL.release();	// Abandon Local PRs and Release Buffers

		if ( ARGS.SWinfo() )  printf("\n");

   	}		// bandoff //


	cclist0  = CCPOOL.getpool();

#ifdef _MonCC2
	CCPOOL.print_CCinfo();
#endif

	lncounts = LNOBJ.createLN(CCPOOL.cccounts,lnlist,cclist0);

	if ( lncounts < 0 ){  fputs("Memory not enough.\n",stderr);  return(1); }
#ifdef _MonLNInfo
	LNOBJ.LNinfo(lncounts,lnlist,cclist0);
#endif
	LNOBJ.stretch(lncounts,lnlist, \
			CCPOOL.cccounts,cclist0,LNLINKPASS,0);

	if ( ! ARGS.SWrjsusp() ){
		if ( 0 == CCOBJ.rejectBox(lnlist,CCPOOL.cccounts,cclist0, \
						BXOBJ.getbuf()) )  break;
	}

	if ( (pass +1) < ARGS.SWrjpass() ){
		for ( ix=0 ; ix<lncounts ; ix++ )  LNOBJ.destroyLN(&lnlist[ix],cclist0);
		for ( ix=0 ; ix < CCPOOL.cccounts ; ix++ ){
			if ( (cclist0[ix].stat & CCSTAT_KILL) != 0 ){
			    CCOBJ.destroyCC_ex(&cclist0[ix],BXOBJ.getbuf());  continue;
			}
		}
	}
    }		// pass //




//------------------------------------- Limit correction phase

	boxcounter = BXOBJ.getboxes();
	bxlst0 = BXOBJ.getbuf();
	for ( ix=0 ; ix<boxcounter ; ix++ )  bxlst0[ix].stat &= ~BXSTAT_REJECT;
	for ( bandoff=0 ; bandoff<width ; bandoff+=BANDWIDTH ){
		if ( ARGS.SWinfo() ){
			printf("------------------------ Band %d ------------------------\n",bandoff/BANDWIDTH);
		}

		boxes = BANDCTL.set_localPRs(bandoff/BANDWIDTH,boxcounter,bxlst0);
		if ( boxes < 0 ){
			fputs("Memory not enough.\n",stderr);
			return(1);
		}
		if ( ARGS.SWinfo() )  printf ("Available boxes = %d\n",boxes);

		cccounts = BANDCTL.createCC();
		cccounts = BANDCTL.checkCC_2();

		if ( ARGS.SWinfo() )  printf("Available CCs = %d\n",cccounts);
		if ( 0 != CCPOOL.addccs(cccounts,BANDCTL.getcclist()) ){
			fprintf(stderr,"Warning: CCs pool overflow\n");
		}

		BANDCTL.release();	// Abandon Local PRs and Release Buffers

		if ( ARGS.SWinfo() )  printf("\n");

   	 }		// bandoff //

#ifdef _MonCC2
	CCPOOL.print_CCinfo();
#endif

	LNOBJ.stretch(lncounts,lnlist, \
			CCPOOL.cccounts,CCPOOL.getpool(),1,1);



//------------------------------------- Export phase


	bxlst0 = BXOBJ.getbuf();

	for ( ix=0 ; ix<lncounts ; ix++ ){
		if ( lnlist[ix].ccs == 0 )  continue;
		if ( (lnlist[ix].stat & LNSTAT_KILL) != 0 ){
			LNOBJ.destroyLN(&lnlist[ix],cclist0);  continue;
		}
		if ( lnlist[ix].avheight2 > LNHLIMIT ){
			LNOBJ.destroyLN(&lnlist[ix],cclist0);  continue;
		}
	}

	STREXP.write(ARGS.outfile(),0,width-1,0,height-1, \
		0,CWID,0,height,lncounts,lnlist,cclist0,bxlst0);

	for ( ix=0 ; ix<lncounts ; ix++ ){
		if ( lnlist[ix].ccs == 0 )  continue;
		if ( (lnlist[ix].stat & LNSTAT_KILL) != 0 ){
			LNOBJ.destroyLN(&lnlist[ix],cclist0);
		}
	}

	for ( ix=0 ; ix < CCPOOL.cccounts ; ix++ ){
		if ( (cclist0[ix].stat & CCSTAT_LINKED) == 0 ){
			CCOBJ.destroyCC_ex(&cclist0[ix],bxlst0);  continue;
		}
		if ( (cclist0[ix].stat & CCSTAT_KILL) != 0 ){
			CCOBJ.destroyCC_ex(&cclist0[ix],bxlst0);  continue;
		}
	}


//	LNOBJ.LNmonitor(lncounts,lnlist,cclist0,bxlst0);


	if ( ARGS.fname_ps() != 0 ){
		PSEXP.writelndata(ARGS.fname_ps(),lncounts,lnlist,cclist0);
	}

	if ( ARGS.fname_apr() != 0 ){
		BXEXP.writebxdata(ARGS.fname_apr(),0,width-1,0,height-1, \
				CWID,CWID,1,height,boxcounter,bxlst0, \
				BXSTAT_ALONE,BXSTAT_ALONE,8);
	}

	if ( ARGS.fname_rpr() != 0 ){
#ifdef	_CreateAlone
		BXEXP.writebxdata(ARGS.fname_rpr(),0,width-1,0,height-1, \
				CWID,CWID,1,height,boxcounter,bxlst0, \
				BXSTAT_ALONE | BXSTAT_CLUSTER,0,8);
#else
		BXEXP.writebxdata(ARGS.fname_rpr(),0,width-1,0,height-1, \
				CWID,CWID,1,height,boxcounter,bxlst0, \
				BXSTAT_CLUSTER,0,8);
#endif
	}

	return(0);
}	
