/*--------------------------------------------------------------
	String Image Viewer  strview  v1.5
		Written  by H.Goto , May  1994
		Modified by H.Goto , Jan. 1995
		Modified by H.Goto , Nov. 1996
		Modified by H.Goto , Aug. 1997
		Modified by H.Goto , July 2000
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-2000  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#define		TITLE		"strview  v1.5"
#define		FONT1		"-adobe-helvetica-bold-r-normal--14-140-75-75-p-82-iso8859-1"
#define		FONT2		"gmw-16"

#define		DEFAULTWIDTH	1024		// Base Window Parameters
#define		DEFAULTHEIGHT	800
#define		DEFAULTX	50
#define		DEFAULTY	50
#define		WW_Max		2048
#define		WH_Max		1600
#define		WW_Min		256
#define		WH_Min		256

#define		STLWIDTH	40		// Status Line Parameters
#define		STLLEFTMARGIN	20
#define		SWLEFTMARGIN	80		// Image Window Parameters
#define		SWRIGHTMARGIN	60
#define		SWTOPMARGIN	(STLWIDTH + 20)
#define		SWBOTTOMMARGIN	20
#define		SWSEP		6
#define		SWFRMWID	2
#define		FNLEFTMARGIN	30


#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	<unistd.h>

#include	"ufilep.h"
#include	"utypes.h"




/*------------------------------------------------------
	Global structure definitions
------------------------------------------------------*/

struct cbox {
	int		x0,y0,w,h;
};


struct FBD {
	int		width;
	int		height;
	int		xoff;
	int		yoff;
	uchar		*buffer;
};




/*------------------------------------------------------
	Bitmap data
------------------------------------------------------*/

#define LRmark_width	16
#define LRmark_height	16
 
static char LRmark_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x02, 0x20, 0x06, 0x30, 0x0a, 0x28, 0x12, 0x24,
   0x22, 0x22, 0x42, 0x21, 0x22, 0x22, 0x12, 0x24, 0x0a, 0x28, 0x06, 0x30,
   0x02, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};




/*------------------------------------------------------
	Window Contorol class
------------------------------------------------------*/

class WINDOWCTL {
    private:
	int		defscrn;
	Visual		*visual;
	int		depth;
	int		wx0,wy0;
	uint		width,height;
	ulong		black,white;
	XImage		*ximage;
	GC		grpgc,txtgc;
	XFontStruct	*font;
	int		fontload(char *fontname);
	uchar		*imgbuf;
    public:
	Display		*display;
	Window		wid;
	int		fontheight;

	int		openwindow(char *display_name);
	void		closewindow(void);
	int		resize(int width,int height);
	int		getwinwidth(void){  return (int)width; }
	int		getwinheight(void){  return (int)height; }
	int		mapwindow(void);
	void		flush(void){ XFlush(display); }
	void		bell(void){ XBell(display,0); }
	int		setximage(uchar *buf,int width,int height);
	XImage		*createximage(uchar *buf,int width,int height);
	Pixmap		createpixmap(uchar *buf,uint width,uint height);
	int		freepixmap(Pixmap pixmap);
	int		releaseximage(void);
	int		putimage(int src_x,int src_y, \
				int dst_x,int dst_y,int width,int height);
	int		putimage(Pixmap pid,int src_x,int src_y, \
				int dst_x,int dst_y,int width,int height);
	int		putimage(XImage *ximage,int src_x,int src_y, \
				int dst_x,int dst_y,int width,int height);
	void		clearall(void);
	void		initscrn(void);
	int		drawrectangle(int x,int y,uint width,uint height,uint lw);
	int		puttext(int x,int y,char *str);
			WINDOWCTL(void);
	virtual		~WINDOWCTL(void);
};


WINDOWCTL :: WINDOWCTL(void){
	display = 0;
	ximage = 0;
	font = 0;
	fontheight = 0;
	wx0 = DEFAULTX;
	wy0 = DEFAULTY;
	width  = DEFAULTWIDTH;
	height = DEFAULTHEIGHT;
	imgbuf = 0;
}


WINDOWCTL :: ~WINDOWCTL(void){
	if ( imgbuf != 0 )  delete []imgbuf;
	if ( ximage != 0 )  XDestroyImage(ximage);
	closewindow();
}


void WINDOWCTL :: closewindow(void){
	if ( display != 0 ){
		XDestroyWindow(display,wid);
		XCloseDisplay(display);  display = 0;
	}
}


int WINDOWCTL :: resize(int width,int height){
	WINDOWCTL :: width  = width;
	WINDOWCTL :: height = height;
	return ( XResizeWindow(display,wid,width,height) );
}


int WINDOWCTL :: fontload(char *fontname){
	XCharStruct	*xcs;
	if ( 0 == (font = XLoadQueryFont(display,fontname)) ){
		fprintf(stderr,"Font  %s  not found.\n",fontname);
		return(-1);
	}
	fontheight = font->max_bounds.ascent;
	return(0);
}


int WINDOWCTL :: openwindow(char *display_name){
	XGCValues	gv;
	XSizeHints	hints;
	XWMHints	wm_hints;
	display = XOpenDisplay(display_name);
	if ( display == 0 )  return(-1);
	defscrn = DefaultScreen(display);
	depth = XDefaultDepth(display,defscrn);
	black = BlackPixel(display,defscrn);
	white = WhitePixel(display,defscrn);
	visual = XDefaultVisual(display,defscrn);
	wid = XCreateSimpleWindow(display,DefaultRootWindow(display), \
					wx0,wy0,width,height,2,black,white);
	XStoreName(display,wid,TITLE);

	hints.flags = PSize | PMinSize | PMaxSize;
	hints.width  = DEFAULTWIDTH;
	hints.height = DEFAULTHEIGHT;
	hints.min_width  = WW_Min;
	hints.min_height = WH_Min;
	hints.max_width  = WW_Max;
	hints.max_height = WH_Max;
	XSetNormalHints(display,wid,&hints);
				       
	gv.foreground = black;
	gv.background = white;
	gv.line_width = SWFRMWID;
	grpgc = XCreateGC(display,wid, \
		(GCForeground | GCBackground | GCLineWidth),&gv);

	if ( 0 != fontload(FONT1) ){
		if ( 0 != fontload(FONT2) )  return(-1);
	}
	gv.font = font->fid;
	txtgc = XCreateGC(display,wid, \
		(GCForeground | GCBackground | GCFont),&gv);

	return(0);
}


int WINDOWCTL :: mapwindow(void){
	XMapWindow(display,wid);
	return(0);
}


int WINDOWCTL :: setximage(uchar *buf,int width,int height){
	int	i,size;
	size = ((width+7)/8) * height;
	if ( ximage != 0 ){
		XDestroyImage(ximage);
	}
	if ( imgbuf != 0 )  delete []imgbuf;
	if ( 0 == (imgbuf = new uchar[size]) )  return(-1);
	for ( i=0 ; i<size ; i++ )  imgbuf[i] = buf[i];
	ximage = XCreateImage(display,visual,1,XYBitmap,0, \
					(char *)imgbuf,width,height,8,0);
	if ( ximage == 0 )  return(-1);
	ximage->byte_order = MSBFirst;
	ximage->bitmap_bit_order = MSBFirst;
	return(0);
}


XImage * WINDOWCTL :: createximage(uchar *buf,int width,int height){
	int	i,size;
	XImage	*ximage;
	size = ((width+7)/8) * height;
	ximage = XCreateImage(display,visual,1,XYBitmap,0, \
					(char *)buf,width,height,8,0);
	if ( ximage != 0 ){
		ximage->byte_order = MSBFirst;
		ximage->bitmap_bit_order = MSBFirst;
	}
	return(ximage);
}


Pixmap WINDOWCTL :: createpixmap(uchar *buf,uint width,uint height){
	return ( XCreatePixmapFromBitmapData(display,wid,(char *)buf,width,height,black,white,depth) );
}


int WINDOWCTL :: freepixmap(Pixmap pixmap){
	return ( XFreePixmap(display,pixmap) );
}


int WINDOWCTL :: putimage(int src_x,int src_y, \
		int dst_x,int dst_y,int width,int height){
	return( XPutImage(display,wid,grpgc,ximage, \
			src_x,src_y,dst_x,dst_y,width,height) );
}


int WINDOWCTL :: putimage(Pixmap pid,int src_x,int src_y, \
		int dst_x,int dst_y,int width,int height){
	return( XCopyArea(display,pid,wid,grpgc,src_x,src_y,width,height,dst_x,dst_y) );
}


int WINDOWCTL :: putimage(XImage *ximage,int src_x,int src_y, \
		int dst_x,int dst_y,int width,int height){
	return( XPutImage(display,wid,grpgc,ximage, \
			src_x,src_y,dst_x,dst_y,width,height) );
}


void WINDOWCTL :: clearall(void){
	XClearWindow(display,wid);
}


void WINDOWCTL :: initscrn(void){
	XSetLineAttributes(display,grpgc,2,LineSolid,CapNotLast,JoinMiter);
	XDrawLine(display,wid,grpgc,0,STLWIDTH -1,width,STLWIDTH -1);
}


int WINDOWCTL :: drawrectangle(int x,int y,uint width,uint height,uint lw){
	XSetLineAttributes(display,grpgc,lw,LineSolid,CapNotLast,JoinMiter);
	return ( XDrawRectangle(display,wid,grpgc,x,y,width,height) );
}


int WINDOWCTL :: puttext(int x,int y,char *str){
	XDrawString(display,wid,txtgc,x,y,str,strlen(str));
	return(0);
}




/*------------------------------------------------------
	Text string Load class
------------------------------------------------------*/

#define		PAGESTRS	64
#define		MAXPAGES	128

class STRLOAD {
    private:
	WINDOWCTL	*WINCTL;
	char		sfname[128];
	char		topdir[256];
	int		page;
	int		pageheight;
	int		pagefiles;
	int		pstart[MAXPAGES+1];
	char		*getfname(int num);
	XImage		*ximage[PAGESTRS];
	uint		width[PAGESTRS];
	uint		height[PAGESTRS];
	int		window_width, window_height;
	int		strwidthlimit, heightlimit;
	XImage		*LRmark;
	int		putstringimage(int fnum);
	int		totalheight(int files);
    public:
	int		Init(void);
	void		setwindowsize(int width,int height);
	void		settopdir(char *dir);
	void		clear(void);
	void		cacheclear(void);
	int		loadpage(int page,int ftype);
	int		loadpagenext(int ftype);
	int		loadpageprev(int ftype);
	int		currentpage(void){  return page;  }
	int		getwidth (int fnum){  return width [fnum];  }
	int		getheight(int fnum){  return height[fnum];  }
	int		putstringimages(void);
	void		writestatus(void);
			STRLOAD(WINDOWCTL *WINCTL);
			~STRLOAD(void);
};


STRLOAD :: STRLOAD(WINDOWCTL *WINCTL){
	int	i;
	STRLOAD :: WINCTL = WINCTL;
	strcpy(topdir,"./");
	page = -1;
	pagefiles = 0;
	for ( i=0 ; i<=MAXPAGES ; i++ )  pstart[i] = 0;
	for ( i=0 ; i<PAGESTRS ; i++ )  ximage[i] = 0;
	LRmark = 0;
}


STRLOAD :: ~STRLOAD(void){
	clear();
	if ( LRmark != 0 )  XDestroyImage(LRmark);
}


int STRLOAD :: Init(void){
	if ( LRmark == 0 ){
		LRmark = WINCTL->createximage((uchar *)LRmark_bits, \
					LRmark_width,LRmark_height);
	}
	if ( LRmark == 0 )  return(-1);
	return(0);
}


void STRLOAD :: setwindowsize(int width,int height){
	window_width  = width;
	window_height = height;
	strwidthlimit = width - SWLEFTMARGIN - SWRIGHTMARGIN;
	heightlimit = height - SWTOPMARGIN - SWBOTTOMMARGIN;
}


void STRLOAD :: settopdir(char *dir){
	strncpy(topdir,dir,248);
	if ( topdir[strlen(topdir) -1] != '/' )  strcat(topdir,"/");
}


void STRLOAD :: clear(void){
	int	i;
	for ( i=0 ; i<PAGESTRS ; i++ ){
		if ( ximage[i] != 0 ){  XDestroyImage(ximage[i]);  ximage[i] = 0; }
	}
}


void STRLOAD :: cacheclear(void){
	int	i;
	clear();
	page = -1;
	pagefiles = 0;
	for ( i=0 ; i<=MAXPAGES ; i++ )  pstart[i] = 0;
}


char * STRLOAD :: getfname(int num){
	sprintf(sfname,"%sSTR%04d",topdir,num);
	return(sfname);
}


int STRLOAD :: loadpage(int page,int ftype){
	int	ret;
	int	sn,start,end;
	uint	w,h,y;
	int	linebytes;
	uchar	*img;
	Pixmap	pid;
	PBMFILE	PBMLD;
	RASLOAD	RASLD;

	start = pstart[page];
	end   = pstart[page+1];
	if ( end <= 0 )  end = start + PAGESTRS;
	if ( start == -1 )  return(-1);

	pageheight = 0;
	pagefiles = 0;
	clear();
	for ( sn=start ; sn<end ; sn++ ){
		if ( ftype == 0 )  ret = PBMLD.open(getfname(sn),"r");
		else               ret = RASLD.fileopen(getfname(sn));
		if ( ret != 0 ){
			pstart[page+1] = -1;
			break;
		}
		if ( ftype == 0 ){
			if ( 0 != PBMLD.getfiletype() ){
				PBMLD.close();
				pstart[page+1] = -1;
				break;
			}
			width [pagefiles] = w = PBMLD.getwidth();
			height[pagefiles] = h = PBMLD.getheight();
		}
		else{	if ( 0 != RASLD.getFiletype() ){
				RASLD.fileclose();
				pstart[page+1] = -1;
				break;
			}
			width [pagefiles] = w = RASLD.getWidth();
			height[pagefiles] = h = RASLD.getHeight();
		}
		if ( (totalheight(pagefiles) + SWSEP + (SWFRMWID *2) + h) > heightlimit ){
			if ( ftype == 0 )  PBMLD.close();
			else               RASLD.fileclose();
			break;
		}

		linebytes = (w + 7)/8;
		if ( 0 != (img = new uchar[linebytes * h +1]) ){
			for ( y=0 ; y<h ; y++ ){
				if ( ftype == 0 ){
					if ( 0 != PBMLD.readline(-1,&img[y * linebytes]) )  break;
				}
				else{	if ( 0 != RASLD.readLine(&img[y * linebytes]) )  break;
				}
			}
			if ( 0 == (ximage[pagefiles] = WINCTL->createximage(img,w,h)) ){
				delete []img;
			}
		}

		pageheight += h;
		++pagefiles;
		if ( ftype == 0 )  PBMLD.close();
		else               RASLD.fileclose();
	}
	if ( pstart[page+1] != -1 )  pstart[page+1] = sn;
	if ( pagefiles == 0 )  pstart[page] = -1;
	if ( STRLOAD :: page == -1 )  STRLOAD :: page = 0;
	return( pagefiles );
}


int STRLOAD :: loadpagenext(int ftype){
	int	ret;
	if ( (page + 1) >= MAXPAGES )  return(-1);
	ret = loadpage(page+1,ftype);
	if ( ret <= 0 )  return(-1);
	++page;
	return(ret);
}


int STRLOAD :: loadpageprev(int ftype){
	int	ret;
	if ( page == 0 )  return(-1);
	ret = loadpage(page-1,ftype);
	if ( ret <= 0 )  return(-1);
	--page;
	return(ret);
}


int STRLOAD :: totalheight(int files){
	int	h,ix;
	if ( pagefiles < files )  files = pagefiles;
	h = 0;
	for ( ix=0 ; ix<files ; ix++ ){
		h += height[ix] + SWFRMWID *2;
	}
	if ( files > 0 )  h += SWSEP * (files -1);
	return(h);
}


int STRLOAD :: putstringimage(int fnum){
	int	y0,w,h,ix;
	int	th,sizeover;
	char	strtmp[32];
	int	mx,my;
	if ( fnum >= pagefiles )  return(-1);
	if ( fnum == 0 )  y0 = SWTOPMARGIN;
	else              y0 = totalheight(fnum) + SWTOPMARGIN + SWSEP;
	w = width[fnum];
	h = height[fnum];
	th = WINCTL->fontheight;
	sizeover = 0;
	if ( w > strwidthlimit ){
		w = strwidthlimit;
		sizeover = -1;
	}
	WINCTL->drawrectangle(SWLEFTMARGIN,y0,w+(SWFRMWID *2),h+(SWFRMWID *2),SWFRMWID);
	sprintf(strtmp,"#%04d",pstart[page] + fnum);
	WINCTL->puttext(FNLEFTMARGIN, y0 + SWFRMWID + ((h + th) /2), strtmp);

	if ( ximage[fnum] != 0 ){
		WINCTL->putimage(ximage[fnum],0,0,SWLEFTMARGIN + SWFRMWID,y0 + SWFRMWID,w,h);
		if ( sizeover ){
			mx = window_width - (SWRIGHTMARGIN + LRmark_width)/2;
			my = y0 + SWFRMWID + ((h - LRmark_height) /2) +1;
			WINCTL->putimage(LRmark,0,0,mx,my,LRmark_width,LRmark_height);
		}
	}

	return(0);
}


int STRLOAD :: putstringimages(void){
	int	ix;
	for ( ix=0 ; ix<pagefiles ; ix++ ){
		if ( putstringimage(ix) != 0 )  return(-1);
	}
	writestatus();
	return(0);
}


void STRLOAD :: writestatus(void){
	char	strtmp[64];
	if ( page >= 0 ){
		if ( pagefiles <= 0 )  sprintf(strtmp,"Page %d :   nothing",page);
		else{	if ( pagefiles == 1 )
				sprintf(strtmp,"Page %d :   %d string ",page,pagefiles);
			else	sprintf(strtmp,"Page %d :   %d strings",page,pagefiles);
			
		}
		WINCTL->puttext(STLLEFTMARGIN,(STLWIDTH + WINCTL->fontheight)/2,strtmp);
	}
}




/*------------------------------------------------------
	Message class
------------------------------------------------------*/

class MSG {
    public:
	void		usage(void);
	void		switcherr(void);
	void		readerr(char *name);
	void		writeerr(char *name);
	void		nonsupported(char *name);
	void		memoryover(void);
	void		filenotfound(void);
};
    	
    	
void MSG :: usage(void){		// Display Usage
	puts("strview  v1.5, Copyright (C) 1994-2000 Hideaki Goto ");
	puts("usage : strview [-options] [top_directory]");
	puts("		-display d : X display to use ");
	puts("		-ras       : read Sun Raster file ");
//	puts("		-help      : print usage ");
//	puts("		-dpi 400   : 400dpi mode (defult: 200dpi)");
}


void MSG :: switcherr(void){
	puts("Switch error.");
}


void MSG :: readerr(char *name){
	printf("Can't read %s.\n",name);
}


void MSG :: writeerr(char *name){
	printf("Can't write %s.\n",name);
}


void MSG :: nonsupported(char *name){
	printf("Non-supported file %s.\n",name);
}


void MSG :: memoryover(void){
	puts("Low memory.");
}


void MSG :: filenotfound(void){
	puts("No file.");
}




/*------------------------------------------------------
	Argument / Option class
------------------------------------------------------*/

class ARGS {
    private:
	char		*display;
	char		*sw_topdir;
    	int		sw_verbose;
	int		sw_filetype;
	int		sw_dpi;
	int		sw_ftype;
	int		rfcount;
    public:
	int		scanargv(int ac,char **av);
	int		SWverbose(void) {  return sw_verbose; }
	int		SWfiletype(void) {  return sw_filetype; }
	int		SWdpi(void) {  return sw_dpi; }
	int		SWftype(void) {  return sw_ftype; }
	char		*topdir(void) {  return sw_topdir; }
	char		*SWdisplay(void){  return display; }
			ARGS(void);
};


ARGS :: ARGS(void){
	sw_topdir = ".";
	sw_verbose = -1;
	sw_filetype = 0;
	sw_dpi = 200;
	sw_ftype = 0;		/* 0:PBM, 1:RAS */
	rfcount = 0;
	display = 0;
}


int ARGS :: scanargv(int ac,char **av){
	int	i1,err,skip;
	char	*sw;
	err = 0;
	skip = 0;
	--ac;
	for( i1 = 0 ; i1 < ac ; ++i1 ){
		if( skip ){  skip = 0;  continue;  }
		if( *( sw = av[i1+1] ) == '-' ){
			++sw;
			if( 0 == strcmp(sw,"display") ){
				if((i1+1)<ac){
					display = av[i1+2];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"dpi") ){
				if((i1+1)<ac){
					sw_dpi = atoi(av[i1+2]);
					if ( sw_dpi != 400 ){  err = -1;  break;  }
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"ras") ){
				sw_ftype = 1;
				continue;
			}
			if( 0 == strcmp(sw,"quiet") ){
				sw_verbose=0;
				continue;
			}
			if( 0 == strcmp(sw,"verbose") ){
				sw_verbose = -1;
				continue;
			}
			if( 0 == strcmp(sw,"help") )  return(1);
			if( 0 == strcmp(sw,"h") )  return(1);
			err = -1;  break;
		}
		else{	sw_topdir = av[i1+1];
		}
	}
	return(err);
}




/*------------------------------------------------------
	Main routine
------------------------------------------------------*/

int main(int ac,char **av){
	ARGS		ARGS;
	MSG		MSG;
	WINDOWCTL	WINCTL;
	STRLOAD		STRLD(&WINCTL);
	XEvent		ev;
	KeySym		key;
	char		kstr[2];
	char		keycode;
	int		Quit;
	int		width0,height0;
	int		retcode,errflag;
	int		pfiles;
	int		exposefirst = -1;
	int		ftype;

	switch ( ARGS.scanargv(ac,av) ){ 
	    case 0:	break;
	    case 1:	MSG.usage();  exit(0);
	    default:	MSG.switcherr();  exit(1);
	}

	ftype = ARGS.SWftype();

	if ( 0 != WINCTL.openwindow(ARGS.SWdisplay()) ){
		exit(1);
	}
	width0  = WINCTL.getwinwidth();
	height0 = WINCTL.getwinheight();
	STRLD.Init();
	STRLD.setwindowsize(width0,height0);
	STRLD.settopdir(ARGS.topdir());

	XSelectInput(WINCTL.display,WINCTL.wid,KeyPressMask | KeyReleaseMask | \
		ExposureMask | ButtonPressMask | ButtonReleaseMask | \
		StructureNotifyMask);

	WINCTL.mapwindow();

	for ( Quit = 0 ; Quit != -1 ; ){
		XNextEvent(WINCTL.display,&ev);
		switch (ev.type){
		    case KeyPress:
				if ( XLookupString((XKeyEvent *)&ev, \
						kstr,2,&key,NULL) == 1 ) ;
				keycode = toupper( *kstr );
				switch(keycode){
				    case 'Q':	Quit = -1;  break;
				    case 'F':
				    case 'N':	pfiles = STRLD.loadpagenext(ftype);
						if ( pfiles >= 0 ){
//						 	printf("Page %d : %d strings.\n",STRLD.currentpage(),pfiles);
							WINCTL.clearall();
							WINCTL.initscrn();
							STRLD.putstringimages();
						}
						else	WINCTL.bell();
						break;
				    case 'B':
				    case 'P':	pfiles = STRLD.loadpageprev(ftype);
						if ( pfiles >= 0 ){
//					 	printf("Page %d : %d strings.\n",STRLD.currentpage(),pfiles);
							WINCTL.clearall();
							WINCTL.initscrn();
							STRLD.putstringimages();
						}
						else	WINCTL.bell();
						break;
				}
				break;
		    case ButtonPress:
				if ( ev.xbutton.button == 3 ){
					Quit = -1;  break;
				}
				break;
		    case Expose:
				if ( exposefirst ){
					pfiles = STRLD.loadpage(0,ftype);
					exposefirst = 0;
					if ( pfiles >= 0 ){
//						printf("Page %d : %d strings.\n",STRLD.currentpage(),pfiles);
					}
					else{	MSG.filenotfound();
					}
				}
				WINCTL.initscrn();
				STRLD.putstringimages();
				break;
		    case ConfigureNotify:
				if ( ev.xconfigure.height != height0 ){
					exposefirst = -1;
				}
				width0  = ev.xconfigure.width;
				height0 = ev.xconfigure.height;
				WINCTL.resize(width0,height0);
				STRLD.setwindowsize(width0,height0);
				WINCTL.initscrn();
				if ( exposefirst ){
					STRLD.cacheclear();
					pfiles = STRLD.loadpage(0,ftype);
					exposefirst = 0;
					if ( pfiles >= 0 ){
//						printf("Page %d : %d strings.\n",STRLD.currentpage(),pfiles);
					}
					else{	MSG.filenotfound();
					}
				}
				else{	STRLD.putstringimages();
				}
				break;
		    default:	break;
		}
	}

	WINCTL.closewindow();

	return(0);
}
